import datetime
import glob
import json
import logging
import os
import pathlib
import shutil
import signal
import sys
from collections import deque
from random import randint, shuffle
from time import time as ttime

import numpy as np
from tqdm import tqdm

import torch
import torch.distributed as dist
import torch.multiprocessing as mp
from torch.nn.parallel import DistributedDataParallel as DDP
from torch.utils.data import DataLoader
from torch.utils.tensorboard import SummaryWriter

# Zluda hijack
import ultimate_rvc.rvc.lib.zluda
from ultimate_rvc.common import TRAINING_MODELS_DIR
from ultimate_rvc.rvc.common import RVC_TRAINING_MODELS_DIR
from ultimate_rvc.rvc.lib.algorithm import commons
from ultimate_rvc.rvc.train.losses import (
    discriminator_loss,
    feature_loss,
    generator_loss,
    kl_loss,
)
from ultimate_rvc.rvc.train.mel_processing import (
    MultiScaleMelSpectrogramLoss,
    mel_spectrogram_torch,
    spec_to_mel_torch,
)
from ultimate_rvc.rvc.train.process.extract_model import extract_model
from ultimate_rvc.rvc.train.utils import (
    HParams,
    latest_checkpoint_path,
    load_checkpoint,
    load_wav_to_torch,
    plot_spectrogram_to_numpy,
    remove_sox_libmso6_from_ld_preload,
    save_checkpoint,
    summarize,
)

logging.getLogger("torch").setLevel(logging.ERROR)
logger = logging.getLogger(__name__)

torch.backends.cudnn.deterministic = False
torch.backends.cudnn.benchmark = True
torch.multiprocessing.set_start_method("spawn", force=True)
os.environ["USE_LIBUV"] = "0" if sys.platform == "win32" else "1"

randomized = True
optimizer = "AdamW"  # "RAdam"
d_lr_coeff = 1.0
g_lr_coeff = 1.0
global_step = 0
lowest_g_value = {"value": float("inf"), "epoch": 0}
lowest_d_value = {"value": float("inf"), "epoch": 0}
consecutive_increases_gen = 0
consecutive_increases_disc = 0

avg_losses = {
    "grad_d_50": deque(maxlen=50),
    "grad_g_50": deque(maxlen=50),
    "disc_loss_50": deque(maxlen=50),
    "fm_loss_50": deque(maxlen=50),
    "kl_loss_50": deque(maxlen=50),
    "mel_loss_50": deque(maxlen=50),
    "gen_loss_50": deque(maxlen=50),
}


class EpochRecorder:
    """
    Records the time elapsed per epoch.
    """

    def __init__(self):
        self.last_time = ttime()

    def record(self):
        """
        Records the elapsed time and returns a formatted string.
        """
        now_time = ttime()
        elapsed_time = now_time - self.last_time
        self.last_time = now_time
        elapsed_time = round(elapsed_time, 1)
        elapsed_time_str = str(datetime.timedelta(seconds=int(elapsed_time)))
        current_time = datetime.datetime.now().strftime("%H:%M:%S")
        return f"time={current_time} | speed={elapsed_time_str}"


def verify_checkpoint_shapes(checkpoint_path: str, model: torch.nn.Module) -> None:
    """
    Verify that the checkpoint and model have the same
    architecture.
    """
    checkpoint = torch.load(checkpoint_path, map_location="cpu", weights_only=False)
    checkpoint_state_dict = checkpoint["model"]
    try:
        if hasattr(model, "module"):
            model_state_dict = model.module.load_state_dict(checkpoint_state_dict)
        else:
            model_state_dict = model.load_state_dict(checkpoint_state_dict)
    except RuntimeError:
        logger.error(  # noqa: TRY400
            "The parameters of the pretrain model such as the sample rate or"
            " architecture do not match the selected model.",
        )
        sys.exit(1)
    else:
        del checkpoint
        del checkpoint_state_dict
        del model_state_dict


def main(
    model_name: str,
    sample_rate: int,
    vocoder: str,
    total_epoch: int,
    batch_size: int,
    save_every_epoch: int,
    save_only_latest: bool,
    save_every_weights: bool,
    pretrain_g: str,
    pretrain_d: str,
    overtraining_detector: bool,
    overtraining_threshold: int,
    cleanup: bool,
    cache_data_in_gpu: bool,
    checkpointing: bool,
    device_type: str,
    gpus: set[int] | None,
) -> None:
    """
    Start the training process.

    Raises:
        RuntimeError: If the sample rate of the pretrained model does not match the dataset audio sample rate.

    """
    remove_sox_libmso6_from_ld_preload()
    experiment_dir = os.path.join(TRAINING_MODELS_DIR, model_name)
    config_save_path = os.path.join(experiment_dir, "config.json")

    # Use a Manager to create a shared list
    manager = mp.Manager()
    global_gen_loss = manager.list([0] * total_epoch)
    global_disc_loss = manager.list([0] * total_epoch)

    with pathlib.Path(config_save_path).open() as f:
        config = json.load(f)
    config = HParams(**config)
    config.data.training_files = os.path.join(experiment_dir, "filelist.txt")

    # Set up distributed training environment for master node.
    # master node is localhost because we are running on a single local
    # machine. master port is randomly selected
    os.environ["MASTER_ADDR"] = "localhost"
    os.environ["MASTER_PORT"] = str(randint(20000, 55555))
    logger.info("MASTER_PORT: %s", os.environ["MASTER_PORT"])
    # Check sample rate
    wavs = glob.glob(
        os.path.join(experiment_dir, "sliced_audios", "*.wav"),
    )
    if wavs:
        _, sr = load_wav_to_torch(wavs[0])
        if sr != sample_rate:
            error_message = (
                f"Error: Pretrained model sample rate ({sample_rate} Hz) does not match"
                f" dataset audio sample rate ({sr} Hz)."
            )
            raise RuntimeError(error_message)
    else:
        logger.warning("No wav file found.")

    device = torch.device(device_type)
    gpus = gpus or {0}
    n_gpus = len(gpus)

    if device.type == "cpu":
        logger.warning("Training with CPU, this will take a long time.")

    def start() -> None:
        """Start the training process with multi-GPU support or CPU."""
        children = []
        pid_data = {"process_pids": []}
        with pathlib.Path(config_save_path).open() as pid_file:
            try:
                existing_data = json.load(pid_file)
                pid_data.update(existing_data)
            except json.JSONDecodeError:
                pass
        with pathlib.Path(config_save_path).open("w") as pid_file:
            for rank, device_id in enumerate(gpus):
                subproc = mp.Process(
                    target=run,
                    args=(
                        rank,
                        n_gpus,
                        experiment_dir,
                        pretrain_g,
                        pretrain_d,
                        total_epoch,
                        save_every_weights,
                        config,
                        device,
                        device_id,
                        model_name,
                        sample_rate,
                        vocoder,
                        batch_size,
                        save_every_epoch,
                        save_only_latest,
                        overtraining_detector,
                        overtraining_threshold,
                        checkpointing,
                        cache_data_in_gpu,
                        global_gen_loss,
                        global_disc_loss,
                    ),
                )
                children.append(subproc)
                subproc.start()
                pid_data["process_pids"].append(subproc.pid)
            json.dump(pid_data, pid_file, indent=4)
        cancel_signal = signal.SIGTERM if os.name == "nt" else -signal.SIGTERM
        error_codes = []
        for i in range(n_gpus):
            children[i].join()
            exit_code = children[i].exitcode
            if exit_code != 0:
                logger.warning(
                    "Process running on device %s exited with code %s.",
                    device_id,
                    exit_code,
                )
                if exit_code != cancel_signal:
                    error_codes.append(exit_code)
        if error_codes:
            err_msg = (
                "One or more training processes failed. See the logs or console for"
                " details."
            )
            raise RuntimeError(err_msg)

    if cleanup:
        logger.info("Removing files from the prior training attempt...")

        # Clean up unnecessary files
        for entry in os.scandir(os.path.join(TRAINING_MODELS_DIR, model_name)):
            if entry.is_file():
                _, file_extension = os.path.splitext(entry.name)
                if file_extension in {".0", ".pth", ".index"}:
                    pathlib.Path(entry.path).unlink()
            elif entry.is_dir() and entry.name == "eval":
                shutil.rmtree(entry.path)

        logger.info("Cleanup done!")
    start()


def run(
    rank,
    n_gpus,
    experiment_dir,
    pretrain_g,
    pretrain_d,
    custom_total_epoch,
    custom_save_every_weights,
    config,
    device,
    device_id,
    model_name,
    sample_rate,
    vocoder,
    batch_size,
    save_every_epoch,
    save_only_latest,
    overtraining_detector,
    overtraining_threshold,
    checkpointing,
    cache_data_in_gpu,
    global_gen_loss,
    global_disc_loss,
):
    """
    Runs the training loop on a specific GPU or CPU.

    Args:
        rank (int): The rank of the current process within the distributed training setup.
        n_gpus (int): The total number of GPUs available for training.
        experiment_dir (str): The directory where experiment logs and checkpoints will be saved.
        pretrain_g (str): Path to the pre-trained generator model.
        pretrain_d (str): Path to the pre-trained discriminator model.
        custom_total_epoch (int): The total number of epochs for training.
        custom_save_every_weights (int): The interval (in epochs) at which to save model weights.
        config (object): Configuration object containing training parameters.
        device (torch.device): The device to use for training (CPU or GPU).

    """
    global global_step, optimizer, lowest_d_value, lowest_g_value, consecutive_increases_gen, consecutive_increases_disc

    if rank == 0:
        writer_eval = SummaryWriter(log_dir=os.path.join(experiment_dir, "eval"))
    else:
        writer_eval = None

    # Initialize distributed training environment for child node.
    dist.init_process_group(
        backend="gloo" if sys.platform == "win32" or device.type != "cuda" else "nccl",
        init_method="env://",
        world_size=n_gpus if device.type == "cuda" else 1,
        rank=rank if device.type == "cuda" else 0,
    )

    torch.manual_seed(config.train.seed)

    if device.type == "cuda":
        torch.cuda.set_device(device_id)

    # Create datasets and dataloaders
    from ultimate_rvc.rvc.train.data_utils import (
        DistributedBucketSampler,
        TextAudioCollateMultiNSFsid,
        TextAudioLoaderMultiNSFsid,
    )

    train_dataset = TextAudioLoaderMultiNSFsid(config.data)
    collate_fn = TextAudioCollateMultiNSFsid()
    train_sampler = DistributedBucketSampler(
        train_dataset,
        batch_size * n_gpus,
        [50, 100, 200, 300, 400, 500, 600, 700, 800, 900],
        num_replicas=n_gpus,
        rank=rank,
        shuffle=True,
    )

    train_loader = DataLoader(
        train_dataset,
        num_workers=4,
        shuffle=False,
        pin_memory=True,
        collate_fn=collate_fn,
        batch_sampler=train_sampler,
        persistent_workers=True,
        prefetch_factor=8,
    )
    if len(train_loader) < 3:
        logger.error(
            "Not enough data in the training set. Perhaps you forgot to slice the"
            " audio files in preprocess?",
        )
        sys.exit(1)
    else:
        g_file = latest_checkpoint_path(experiment_dir, "G_*.pth")
        if g_file != None:
            logger.info("Checking saved weights...")
            g = torch.load(g_file, map_location="cpu", weights_only=False)
            if (
                optimizer == "RAdam"
                and "amsgrad" in g["optimizer"]["param_groups"][0].keys()
            ):
                optimizer = "AdamW"
                logger.info(
                    "Optimizer choice has been reverted to %s to match the saved D/G"
                    " weights.",
                    optimizer,
                )
            elif (
                optimizer == "AdamW"
                and "decoupled_weight_decay" in g["optimizer"]["param_groups"][0].keys()
            ):
                optimizer = "RAdam"
                logger.info(
                    "Optimizer choice has been reverted to %s to match the saved D/G"
                    " weights.",
                    optimizer,
                )
            del g

    # Initialize models and optimizers
    from ultimate_rvc.rvc.lib.algorithm.discriminators import MultiPeriodDiscriminator
    from ultimate_rvc.rvc.lib.algorithm.synthesizers import Synthesizer

    # NOTE checkingpointing here means whether or not activations are
    # saved during forward pass for backpropagation during backward pass

    net_g = Synthesizer(
        config.data.filter_length // 2 + 1,
        config.train.segment_size // config.data.hop_length,
        **config.model,
        use_f0=True,
        sr=sample_rate,
        vocoder=vocoder,
        checkpointing=checkpointing,
        randomized=randomized,
    )

    net_d = MultiPeriodDiscriminator(
        config.model.use_spectral_norm,
        checkpointing=checkpointing,
    )

    if device.type == "cuda":
        net_g = net_g.cuda(device_id)
        net_d = net_d.cuda(device_id)
    else:
        net_g = net_g.to(device)
        net_d = net_d.to(device)

    if optimizer == "AdamW":
        optimizer = torch.optim.AdamW
    elif optimizer == "RAdam":
        optimizer = torch.optim.RAdam

    optim_g = optimizer(
        net_g.parameters(),
        config.train.learning_rate * g_lr_coeff,
        betas=config.train.betas,
        eps=config.train.eps,
    )
    optim_d = optimizer(
        net_d.parameters(),
        config.train.learning_rate * d_lr_coeff,
        betas=config.train.betas,
        eps=config.train.eps,
    )

    fn_mel_loss = MultiScaleMelSpectrogramLoss(sample_rate=sample_rate)

    # Wrap models with DDP for multi-gpu processing
    if n_gpus > 1 and device.type == "cuda":
        net_g = DDP(net_g, device_ids=[device_id])
        net_d = DDP(net_d, device_ids=[device_id])

    # Load checkpoint if available
    try:
        logger.info("Starting training...")
        _, _, _, epoch_str, lowest_d_value, consecutive_increases_disc = (
            load_checkpoint(
                latest_checkpoint_path(experiment_dir, "D_*.pth"),
                net_d,
                optim_d,
            )
        )
        _, _, _, epoch_str, lowest_g_value, consecutive_increases_gen = load_checkpoint(
            latest_checkpoint_path(experiment_dir, "G_*.pth"),
            net_g,
            optim_g,
        )
        epoch_str += 1
        global_step = (epoch_str - 1) * len(train_loader)

    except Exception:
        epoch_str = 1
        global_step = 0
        if pretrain_g not in {"", "None"}:
            if rank == 0:
                verify_checkpoint_shapes(pretrain_g, net_g)
                logger.info("Loaded pretrained (G) '%s'", pretrain_g)
            if hasattr(net_g, "module"):
                net_g.module.load_state_dict(
                    torch.load(pretrain_g, map_location="cpu", weights_only=False)[
                        "model"
                    ],
                )
            else:
                net_g.load_state_dict(
                    torch.load(pretrain_g, map_location="cpu", weights_only=False)[
                        "model"
                    ],
                )

        if pretrain_d not in {"", "None"}:
            if rank == 0:
                verify_checkpoint_shapes(pretrain_d, net_d)
                logger.info("Loaded pretrained (D) '%s'", pretrain_d)
            if hasattr(net_d, "module"):
                net_d.module.load_state_dict(
                    torch.load(pretrain_d, map_location="cpu", weights_only=False)[
                        "model"
                    ],
                )
            else:
                net_d.load_state_dict(
                    torch.load(pretrain_d, map_location="cpu", weights_only=False)[
                        "model"
                    ],
                )

    # Initialize schedulers
    scheduler_g = torch.optim.lr_scheduler.ExponentialLR(
        optim_g,
        gamma=config.train.lr_decay,
        last_epoch=epoch_str - 2,
    )
    scheduler_d = torch.optim.lr_scheduler.ExponentialLR(
        optim_d,
        gamma=config.train.lr_decay,
        last_epoch=epoch_str - 2,
    )

    cache = []
    # get the first sample as reference for tensorboard evaluation
    # custom reference temporarily disabled
    if (
        True == False
        and pathlib.Path(
            os.path.join(RVC_TRAINING_MODELS_DIR, "reference", f"ref{sample_rate}.wav")
        ).is_file()
    ):
        phone = np.load(
            os.path.join(
                RVC_TRAINING_MODELS_DIR,
                "reference",
                f"ref{sample_rate}_feats.npy",
            ),
        )
        # expanding x2 to match pitch size
        phone = np.repeat(phone, 2, axis=0)
        phone = torch.FloatTensor(phone).unsqueeze(0).to(device)
        phone_lengths = torch.LongTensor(phone.size(0)).to(device)
        pitch = np.load(
            os.path.join(
                RVC_TRAINING_MODELS_DIR,
                "reference",
                f"ref{sample_rate}_f0c.npy",
            ),
        )
        # removed last frame to match features
        pitch = torch.LongTensor(pitch[:-1]).unsqueeze(0).to(device)
        pitchf = np.load(
            os.path.join(
                RVC_TRAINING_MODELS_DIR,
                "reference",
                f"ref{sample_rate}_f0f.npy",
            ),
        )
        # removed last frame to match features
        pitchf = torch.FloatTensor(pitchf[:-1]).unsqueeze(0).to(device)
        sid = torch.LongTensor([0]).to(device)
        reference = (
            phone,
            phone_lengths,
            pitch,
            pitchf,
            sid,
        )
    else:
        for info in train_loader:
            phone, phone_lengths, pitch, pitchf, _, _, _, _, sid = info
            if device.type == "cuda":
                reference = (
                    phone.cuda(device_id, non_blocking=True),
                    phone_lengths.cuda(device_id, non_blocking=True),
                    pitch.cuda(device_id, non_blocking=True),
                    pitchf.cuda(device_id, non_blocking=True),
                    sid.cuda(device_id, non_blocking=True),
                )
            else:
                reference = (
                    phone.to(device),
                    phone_lengths.to(device),
                    pitch.to(device),
                    pitchf.to(device),
                    sid.to(device),
                )
            break

    if epoch_str > custom_total_epoch:
        cleanup_training_processes(experiment_dir)
    for epoch in range(epoch_str, custom_total_epoch + 1):
        train_and_evaluate(
            rank,
            epoch,
            config,
            [net_g, net_d],
            [optim_g, optim_d],
            [scheduler_g, scheduler_d],
            [train_loader, None],
            [writer_eval],
            cache,
            custom_save_every_weights,
            custom_total_epoch,
            device,
            device_id,
            reference,
            fn_mel_loss,
            model_name,
            experiment_dir,
            sample_rate,
            vocoder,
            save_every_epoch,
            save_only_latest,
            overtraining_detector,
            overtraining_threshold,
            cache_data_in_gpu,
            global_gen_loss,
            global_disc_loss,
        )


def train_and_evaluate(
    rank,
    epoch,
    config,
    nets,
    optims,
    schedulers,
    loaders,
    writers,
    cache,
    custom_save_every_weights,
    custom_total_epoch,
    device,
    device_id,
    reference,
    fn_mel_loss,
    model_name,
    experiment_dir,
    sample_rate,
    vocoder,
    save_every_epoch,
    save_only_latest,
    overtraining_detector,
    overtraining_threshold,
    cache_data_in_gpu,
    global_gen_loss,
    global_disc_loss,
) -> None:
    """Train and evaluates the model for one epoch."""
    global global_step, lowest_g_value, lowest_d_value, consecutive_increases_gen, consecutive_increases_disc

    model_add = []
    checkpoint_idxs = []
    done = False

    net_g, net_d = nets
    optim_g, optim_d = optims
    scheduler_g, scheduler_d = schedulers
    skip_g_scheduler, skip_d_scheduler = False, False
    train_loader = loaders[0] if loaders is not None else None
    if writers is not None:
        writer = writers[0]

    train_loader.batch_sampler.set_epoch(epoch)

    net_g.train()
    net_d.train()

    # Data caching
    if device.type == "cuda" and cache_data_in_gpu:
        if cache == []:
            for batch_idx, info in enumerate(train_loader):
                # phone, phone_lengths, pitch, pitchf, spec, spec_lengths, wave, wave_lengths, sid
                info = [tensor.cuda(device_id, non_blocking=True) for tensor in info]
                cache.append((batch_idx, info))
        shuffle(cache)
        data_iterator = cache
    else:
        data_iterator = enumerate(train_loader)

    epoch_recorder = EpochRecorder()
    with tqdm(total=len(train_loader), leave=False) as pbar:
        for batch_idx, info in data_iterator:
            if device.type == "cuda" and not cache_data_in_gpu:
                info = [tensor.cuda(device_id, non_blocking=True) for tensor in info]
            elif device.type != "cuda":
                info = [tensor.to(device) for tensor in info]
            # else iterator is going thru a cached list with a device already assigned

            (
                phone,
                phone_lengths,
                pitch,
                pitchf,
                spec,
                spec_lengths,
                wave,
                wave_lengths,
                sid,
            ) = info

            # Forward pass
            model_output = net_g(
                phone,
                phone_lengths,
                pitch,
                pitchf,
                spec,
                spec_lengths,
                sid,
            )
            y_hat, ids_slice, x_mask, z_mask, (z, z_p, m_p, logs_p, m_q, logs_q) = (
                model_output
            )
            # slice of the original waveform to match a generate slice
            if randomized:
                wave = commons.slice_segments(
                    wave,
                    ids_slice * config.data.hop_length,
                    config.train.segment_size,
                    dim=3,
                )
            y_d_hat_r, y_d_hat_g, _, _ = net_d(wave, y_hat.detach())
            loss_disc, _, _ = discriminator_loss(y_d_hat_r, y_d_hat_g)
            # Discriminator backward and update
            global_disc_loss[epoch - 1] += loss_disc.item()
            optim_d.zero_grad()
            loss_disc.backward()
            grad_norm_d = commons.grad_norm(net_d.parameters())
            optim_d.step()

            # Generator backward and update
            _, y_d_hat_g, fmap_r, fmap_g = net_d(wave, y_hat)
            loss_mel = fn_mel_loss(wave, y_hat) * config.train.c_mel / 3.0
            loss_kl = kl_loss(z_p, logs_q, m_p, logs_p, z_mask) * config.train.c_kl
            loss_fm = feature_loss(fmap_r, fmap_g)
            loss_gen, _ = generator_loss(y_d_hat_g)
            loss_gen_all = loss_gen + loss_fm + loss_mel + loss_kl
            global_gen_loss[epoch - 1] += loss_gen_all.item()
            optim_g.zero_grad()
            loss_gen_all.backward()
            grad_norm_g = commons.grad_norm(net_g.parameters())
            optim_g.step()

            global_step += 1

            # queue for rolling losses over 50 steps
            avg_losses["grad_d_50"].append(grad_norm_d)
            avg_losses["grad_g_50"].append(grad_norm_g)
            avg_losses["disc_loss_50"].append(loss_disc.detach())
            avg_losses["fm_loss_50"].append(loss_fm.detach())
            avg_losses["kl_loss_50"].append(loss_kl.detach())
            avg_losses["mel_loss_50"].append(loss_mel.detach())
            avg_losses["gen_loss_50"].append(loss_gen_all.detach())

            if rank == 0 and global_step % 50 == 0:
                # logging rolling averages
                scalar_dict = {
                    "grad_avg_50/norm_d": (
                        sum(avg_losses["grad_d_50"]) / len(avg_losses["grad_d_50"])
                    ),
                    "grad_avg_50/norm_g": (
                        sum(avg_losses["grad_g_50"]) / len(avg_losses["grad_g_50"])
                    ),
                    "loss_avg_50/d/total": torch.mean(
                        torch.stack(list(avg_losses["disc_loss_50"])),
                    ),
                    "loss_avg_50/g/fm": torch.mean(
                        torch.stack(list(avg_losses["fm_loss_50"])),
                    ),
                    "loss_avg_50/g/kl": torch.mean(
                        torch.stack(list(avg_losses["kl_loss_50"])),
                    ),
                    "loss_avg_50/g/mel": torch.mean(
                        torch.stack(list(avg_losses["mel_loss_50"])),
                    ),
                    "loss_avg_50/g/total": torch.mean(
                        torch.stack(list(avg_losses["gen_loss_50"])),
                    ),
                }
                summarize(
                    writer=writer,
                    global_step=global_step,
                    scalars=scalar_dict,
                )

            pbar.update(1)
        # end of batch train
    # end of tqdm
    scheduler_d.step()
    scheduler_g.step()

    with torch.no_grad():
        torch.cuda.empty_cache()
    # Logging and checkpointing
    if rank == 0:
        avg_global_disc_loss = global_disc_loss[epoch - 1] / len(train_loader.dataset)
        avg_global_gen_loss = global_gen_loss[epoch - 1] / len(train_loader.dataset)

        min_delta = 0.004

        if avg_global_disc_loss < lowest_d_value["value"] - min_delta:
            lowest_d_value = {"value": avg_global_disc_loss, "epoch": epoch}
            consecutive_increases_disc = 0
        else:
            consecutive_increases_disc += 1

        if avg_global_gen_loss < lowest_g_value["value"] - min_delta:
            logger.info(
                "New best epoch %d with average generator loss %.3f and discriminator"
                " loss %.3f",
                epoch,
                avg_global_gen_loss,
                avg_global_disc_loss,
            )
            lowest_g_value = {"value": avg_global_gen_loss, "epoch": epoch}
            consecutive_increases_gen = 0
            model_add.append(
                os.path.join(experiment_dir, f"{model_name}_best.pth"),
            )
        else:
            consecutive_increases_gen += 1

        # used for tensorboard chart - all/mel
        mel = spec_to_mel_torch(
            spec,
            config.data.filter_length,
            config.data.n_mel_channels,
            config.data.sample_rate,
            config.data.mel_fmin,
            config.data.mel_fmax,
        )
        # used for tensorboard chart - slice/mel_org
        if randomized:
            y_mel = commons.slice_segments(
                mel,
                ids_slice,
                config.train.segment_size // config.data.hop_length,
                dim=3,
            )
        else:
            y_mel = mel
        # used for tensorboard chart - slice/mel_gen
        y_hat_mel = mel_spectrogram_torch(
            y_hat.float().squeeze(1),
            config.data.filter_length,
            config.data.n_mel_channels,
            config.data.sample_rate,
            config.data.hop_length,
            config.data.win_length,
            config.data.mel_fmin,
            config.data.mel_fmax,
        )

        lr = optim_g.param_groups[0]["lr"]

        scalar_dict = {
            "loss/g/total": loss_gen_all,
            "loss/d/total": loss_disc,
            "learning_rate": lr,
            "grad/norm_d": grad_norm_d,
            "grad/norm_g": grad_norm_g,
            "loss/g/fm": loss_fm,
            "loss/g/mel": loss_mel,
            "loss/g/kl": loss_kl,
        }

        image_dict = {
            "slice/mel_org": plot_spectrogram_to_numpy(y_mel[0].data.cpu().numpy()),
            "slice/mel_gen": plot_spectrogram_to_numpy(y_hat_mel[0].data.cpu().numpy()),
            "all/mel": plot_spectrogram_to_numpy(mel[0].data.cpu().numpy()),
        }
        overtrain_info = ""
        # Print training progress
        lowest_g_value_rounded = float(lowest_g_value["value"])
        lowest_g_value_rounded = round(lowest_g_value_rounded, 3)

        record = f"{model_name} | epoch={epoch} | {epoch_recorder.record()}"
        record += (
            f" | best avg-gen-loss={lowest_g_value_rounded:.3f} (epoch"
            f" {lowest_g_value['epoch']})"
        )
        # Check overtraining
        if overtraining_detector:
            overtrain_info = (
                f"Average epoch generator loss {avg_global_gen_loss:.3f} and"
                f" discriminator loss {avg_global_disc_loss:.3f}"
            )

            remaining_epochs_gen = max(
                overtraining_threshold - consecutive_increases_gen,
                0,
            )
            remaining_epochs_disc = max(
                overtraining_threshold * 2 - consecutive_increases_disc,
                0,
            )
            record += (
                " | overtrain countdown: g="
                f"{remaining_epochs_gen},d={remaining_epochs_disc} |"
                f" avg-gen-loss={avg_global_gen_loss:.3f} | avg-"
                f"disc-loss={avg_global_disc_loss:.3f}"
            )

            if remaining_epochs_disc == 0 or remaining_epochs_gen == 0:
                record += (
                    f"\nOvertraining detected at epoch {epoch} with average"
                    f" generator loss {avg_global_gen_loss:.3f} and discriminator loss"
                    f" {avg_global_disc_loss:.3f}"
                )
                done = True
        print(record)

        # Save weights, checkpoints and reference inference results
        # every N epochs
        if epoch % save_every_epoch == 0:
            with torch.no_grad():
                if hasattr(net_g, "module"):
                    o, *_ = net_g.module.infer(*reference)
                else:
                    o, *_ = net_g.infer(*reference)
            audio_dict = {f"gen/audio_{global_step:07d}": o[0, :, :]}
            summarize(
                writer=writer,
                global_step=global_step,
                images=image_dict,
                scalars=scalar_dict,
                audios=audio_dict,
                audio_sample_rate=config.data.sample_rate,
            )
            checkpoint_idxs.append(2333333)
            if not save_only_latest:
                checkpoint_idxs.append(epoch)

            if custom_save_every_weights:
                model_add.append(
                    os.path.join(experiment_dir, f"{model_name}_{epoch}.pth"),
                )
        else:
            summarize(
                writer=writer,
                global_step=global_step,
                images=image_dict,
                scalars=scalar_dict,
            )
        for idx in checkpoint_idxs:
            save_checkpoint(
                net_g,
                optim_g,
                config.train.learning_rate,
                epoch,
                lowest_g_value,
                consecutive_increases_gen,
                os.path.join(experiment_dir, f"G_{idx}.pth"),
            )
            save_checkpoint(
                net_d,
                optim_d,
                config.train.learning_rate,
                epoch,
                lowest_d_value,
                consecutive_increases_disc,
                os.path.join(experiment_dir, f"D_{idx}.pth"),
            )
        if model_add:
            ckpt = (
                net_g.module.state_dict()
                if hasattr(net_g, "module")
                else net_g.state_dict()
            )
            for m in model_add:
                extract_model(
                    ckpt=ckpt,
                    sr=sample_rate,
                    name=model_name,
                    model_dir=m,
                    epoch=epoch,
                    step=global_step,
                    hps=config,
                    overtrain_info=overtrain_info,
                    vocoder=vocoder,
                )
        # Check completion
        if epoch >= custom_total_epoch:
            lowest_g_value_rounded = float(lowest_g_value["value"])
            lowest_g_value_rounded = round(lowest_g_value_rounded, 3)
            print(
                f"Training has been successfully completed with {epoch} epoch(s),"
                f" {global_step} step(s) and {round(avg_global_gen_loss, 3)} average"
                " generator loss.",
            )
            print(
                f"Lowest average generator loss: {lowest_g_value_rounded} at epoch"
                f" {lowest_g_value['epoch']}",
            )

            done = True
        with torch.no_grad():
            torch.cuda.empty_cache()
        if done:
            cleanup_training_processes(experiment_dir)
            os._exit(0)


def cleanup_training_processes(experiment_dir) -> None:
    dist.destroy_process_group()
    pid_file_path = os.path.join(experiment_dir, "config.json")
    with pathlib.Path(pid_file_path).open() as pid_file:
        pid_data = json.load(pid_file)
    with pathlib.Path(pid_file_path).open("w") as pid_file:
        pid_data.pop("process_pids", None)
        json.dump(pid_data, pid_file, indent=4)



"""Extra typing for the Ultimate RVC project."""

from __future__ import annotations

from collections.abc import Mapping, Sequence
from enum import IntEnum, StrEnum
from os import PathLike

type StrPath = str | PathLike[str]

type Json = Mapping[str, Json] | Sequence[Json] | str | int | float | bool | None


class SeparationModel(StrEnum):
    """Enumeration of audio separation models."""

    UVR_MDX_NET_VOC_FT = "UVR-MDX-NET-Voc_FT.onnx"
    UVR_MDX_NET_KARA_2 = "UVR_MDXNET_KARA_2.onnx"
    REVERB_HQ_BY_FOXJOY = "Reverb_HQ_By_FoxJoy.onnx"


class SegmentSize(IntEnum):
    """Enumeration of segment sizes for audio separation."""

    SEG_64 = 64
    SEG_128 = 128
    SEG_256 = 256
    SEG_512 = 512
    SEG_1024 = 1024
    SEG_2048 = 2048


class F0Method(StrEnum):
    """Enumeration of pitch extraction methods."""

    RMVPE = "rmvpe"
    CREPE = "crepe"
    CREPE_TINY = "crepe-tiny"
    FCPE = "fcpe"


class EmbedderModel(StrEnum):
    """Enumeration of audio embedding models."""

    CONTENTVEC = "contentvec"
    CHINESE_HUBERT_BASE = "chinese-hubert-base"
    JAPANESE_HUBERT_BASE = "japanese-hubert-base"
    KOREAN_HUBERT_BASE = "korean-hubert-base"
    CUSTOM = "custom"


class RVCContentType(StrEnum):
    """Enumeration of valid content to convert with RVC."""

    VOCALS = "vocals"
    VOICE = "voice"
    SPEECH = "speech"
    AUDIO = "audio"


class SampleRate(IntEnum):
    """Enumeration of supported audio sample rates."""

    HZ_16000 = 16000
    HZ_44100 = 44100
    HZ_48000 = 48000
    HZ_96000 = 96000
    HZ_192000 = 192000


class AudioExt(StrEnum):
    """Enumeration of supported audio file formats."""

    MP3 = "mp3"
    WAV = "wav"
    FLAC = "flac"
    OGG = "ogg"
    M4A = "m4a"
    AAC = "aac"


class DeviceType(StrEnum):
    """Enumeration of device types for training voice models."""

    AUTOMATIC = "Automatic"
    CPU = "CPU"
    GPU = "GPU"


class TrainingSampleRate(StrEnum):
    """Enumeration of sample rates for training voice models."""

    HZ_32K = "32000"
    HZ_40K = "40000"
    HZ_48K = "48000"


class PretrainedSampleRate(StrEnum):
    """Enumeration of valid sample rates for pretrained models."""

    HZ_32K = "32k"
    HZ_40K = "40k"
    HZ_44K = "44k"
    HZ_48K = "48k"


class TrainingF0Method(StrEnum):
    """Enumeration of pitch extraction methods for training."""

    RMVPE = "rmvpe"
    CREPE = "crepe"
    CREPE_TINY = "crepe-tiny"


class AudioSplitMethod(StrEnum):
    """
    Enumeration of methods to use for splitting audio files during
    dataset preprocessing.
    """

    SKIP = "Skip"
    SIMPLE = "Simple"
    AUTOMATIC = "Automatic"


class Vocoder(StrEnum):
    """Enumeration of vocoders for training voice models."""

    HIFI_GAN = "HiFi-GAN"
    MRF_HIFI_GAN = "MRF HiFi-GAN"
    REFINE_GAN = "RefineGAN"


class IndexAlgorithm(StrEnum):
    """Enumeration of indexing algorithms for training voice models."""

    AUTO = "Auto"
    FAISS = "Faiss"
    KMEANS = "KMeans"


class PretrainedType(StrEnum):
    """
    Enumeration of the possible types of pretrained models to finetune
    voice models on.
    """

    NONE = "None"
    DEFAULT = "Default"
    CUSTOM = "Custom"



"""Module which defines the code for the "Settings" tab."""

from __future__ import annotations

from typing import TYPE_CHECKING

from functools import partial

import gradio as gr

from ultimate_rvc.core.manage.config import (
    delete_all_configs,
    delete_configs,
    get_config_names,
)
from ultimate_rvc.core.manage.settings import delete_temp_files
from ultimate_rvc.web.common import (
    confirm_box_js,
    confirmation_harness,
    exception_harness,
    load_total_config_values,
    render_msg,
    save_total_config_values,
    setup_delete_event,
    update_dropdowns,
)

if TYPE_CHECKING:
    from ultimate_rvc.web.config.main import SettingsManagementConfig, TotalConfig


def render(total_config: TotalConfig) -> None:
    """
    Render "Settings" tab.

    Parameters
    ----------
    total_config : TotalConfig
        The total configuration object containing all the settings.

    """
    tab_config = total_config.management.settings
    tab_config.dummy_checkbox.instantiate()
    _render_config_files_tab(tab_config, total_config)
    _render_temp_files_tab(tab_config)


def _render_config_files_tab(
    tab_config: SettingsManagementConfig,
    total_config: TotalConfig,
) -> None:
    components = [config.instance for config in total_config.all]
    with gr.Tab("Configuration files"):
        with gr.Accordion("Save configuration"), gr.Row():
            with gr.Column():
                save_config_name = gr.Textbox(
                    label="Configuration name",
                    info=(
                        "The name of the configuration to save the current UI"
                        " settings to."
                    ),
                    placeholder="Enter a name for the configuration",
                    max_lines=1,
                    value="Default Configuration",
                )
                save_config_btn = gr.Button("Save", variant="primary")
            with gr.Column():
                save_config_msg = gr.Textbox(label="Output message", interactive=False)

        with gr.Accordion("Load configuration", open=False), gr.Row():
            with gr.Column():
                tab_config.load_config_name.instance.render()
                load_config_btn = gr.Button("Load", variant="primary")
            with gr.Column():
                load_config_msg = gr.Textbox(label="Output message", interactive=False)

        save_config_btn.click(
            exception_harness(save_total_config_values),
            inputs=[save_config_name, *components],
            outputs=save_config_msg,
        ).success(
            partial(render_msg, "[-] Successfully saved configuration!"),
            outputs=save_config_msg,
            show_progress="hidden",
        ).then(
            partial(update_dropdowns, get_config_names, 1, value_indices=[0]),
            inputs=save_config_name,
            outputs=tab_config.load_config_name.instance,
            show_progress="hidden",
        ).then(
            partial(update_dropdowns, get_config_names, 1, [], [0]),
            outputs=tab_config.delete_config_names.instance,
            show_progress="hidden",
        )

        load_config_btn.click(
            exception_harness(load_total_config_values),
            inputs=tab_config.load_config_name.instance,
            outputs=components,
            show_progress_on=load_config_msg,
        ).success(
            partial(render_msg, "[-] Successfully loaded configuration!"),
            outputs=load_config_msg,
            show_progress="hidden",
        )

        with gr.Accordion("Delete configuration(s)", open=False), gr.Row():
            with gr.Column():
                tab_config.delete_config_names.instance.render()
                delete_config_btn = gr.Button("Delete selected", variant="secondary")
                delete_all_config_btn = gr.Button("Delete all", variant="primary")
            with gr.Column():
                delete_config_msg = gr.Textbox(
                    label="Output message",
                    interactive=False,
                )
        delete_config_click = setup_delete_event(
            delete_config_btn,
            delete_configs,
            [
                tab_config.dummy_checkbox.instance,
                tab_config.delete_config_names.instance,
            ],
            delete_config_msg,
            "Are you sure you want to delete the configurations with the selected"
            " names?",
            "[-] Successfully deleted the configurations with the selected names!",
        )
        delete_all_config_click = setup_delete_event(
            delete_all_config_btn,
            delete_all_configs,
            [tab_config.dummy_checkbox.instance],
            delete_config_msg,
            "Are you sure you want to delete all configuration?",
            "[-] Successfully deleted all configuration!",
        )
        for event in [delete_config_click, delete_all_config_click]:
            event.success(
                partial(update_dropdowns, get_config_names, 2, [], [1]),
                outputs=[
                    tab_config.load_config_name.instance,
                    tab_config.delete_config_names.instance,
                ],
                show_progress="hidden",
            )


def _render_temp_files_tab(tab_config: SettingsManagementConfig) -> None:
    with gr.Tab("Temporary files"):
        gr.Markdown("")
        with gr.Row(equal_height=True):
            temporary_files_btn = gr.Button("Delete all", variant="primary")
            temporary_files_msg = gr.Textbox(label="Output message", interactive=False)

        temporary_files_btn.click(
            confirmation_harness(delete_temp_files),
            inputs=tab_config.dummy_checkbox.instance,
            outputs=temporary_files_msg,
            js=confirm_box_js(
                "Are you sure you want to delete all temporary files? Any files"
                " uploaded directly via the UI will not be available for further"
                " processing until they are re-uploaded.",
            ),
        ).success(
            partial(
                render_msg,
                "[-] Successfully deleted all temporary files!",
            ),
            outputs=temporary_files_msg,
            show_progress="hidden",
        )



"""
Module which defines extra types used by modules in the
ultimate_rvc.core.generate package.
"""

from __future__ import annotations

from typing import Literal

from enum import StrEnum, auto

from pydantic import BaseModel, ConfigDict

# NOTE these types are used at runtime by pydantic so cannot be
# relegated to a IF TYPE_CHECKING block
from ultimate_rvc.typing_extra import AudioExt, EmbedderModel, F0Method  # noqa: TC002


class SongSourceType(StrEnum):
    """The type of source providing the song to generate a cover of."""

    URL = auto()
    FILE = auto()
    SONG_DIR = auto()


class AudioExtInternal(StrEnum):
    """Audio file formats for internal use."""

    MP3 = "mp3"
    WAV = "wav"
    FLAC = "flac"
    OGG = "ogg"
    IPOD = "ipod"
    ADTS = "adts"


class DirectoryMetaData(BaseModel):
    """
    Metadata for a directory.

    Attributes
    ----------
    name : str
        The name of the directory.
    path : str
        The path of the directory.

    """

    name: str
    path: str


class FileMetaData(BaseModel):
    """
    Metadata for a file.

    Attributes
    ----------
    name : str
        The name of the file.
    hash_id : str
        The hash ID of the file.

    """

    name: str
    hash_id: str


class WaveifiedAudioMetaData(BaseModel):
    """
    Metadata for a waveified audio track.

    Attributes
    ----------
    audio_track : FileMetaData
        Metadata for the audio track that was waveified.

    """

    audio_track: FileMetaData


class SeparatedAudioMetaData(BaseModel):
    """
    Metadata for a separated audio track.

    Attributes
    ----------
    audio_track : FileMetaData
        Metadata for the audio track that was separated.
    model_name : str
        The name of the model used for separation.
    segment_size : int
        The segment size used for separation.

    """

    audio_track: FileMetaData
    model_name: str
    segment_size: int

    model_config = ConfigDict(protected_namespaces=())


class RVCAudioMetaData(BaseModel):
    """
    Metadata for a voice converted audio track.

    Attributes
    ----------
    voice_track : FileMetaData
        Metadata for the audio track that was voice converted.
    model_name : str
        The name of the model used for voice conversion.
    n_semitones : int
        The number of semitones the converted audio was pitch-shifted
        by.
    f0_methods : list[F0Method]
        The methods used for pitch extraction.
    index_rate : float
        The influence of the index file on the voice conversion.
    rms_mix_rate : float
        The blending rate of the volume envelope of the converted
        audio.
    protect_rate : float
        The protection rate for consonants and breathing sounds used
        for the audio conversion.
    hop_length : int
        The hop length used for CREPE-based pitch extraction.
    split_audio : bool
        Whether the audio track was split before it was converted.
    autotune_audio : bool
        Whether autotune was applied to the converted audio.
    autotune_strength : float
        The strength of the autotune effect applied to the converted
        audio.
    clean_audio : bool
        Whether the converted audio was cleaned.
    clean_strength : float
        The intensity of the cleaning that was applied to the converted
        audio.
    embedder_model : EmbedderModel
        The model used for generating speaker embeddings.
    custom_embedder_model : str | None
        The name of a custom embedder model used for generating speaker
        embeddings.
    sid : int
        The speaker id used for multi-speaker conversion.

    """

    audio_track: FileMetaData
    model_name: str
    n_semitones: int
    f0_methods: list[F0Method]
    index_rate: float
    rms_mix_rate: float
    protect_rate: float
    hop_length: int
    split_audio: bool
    autotune_audio: bool
    autotune_strength: float
    clean_audio: bool
    clean_strength: float
    embedder_model: EmbedderModel
    custom_embedder_model: str | None
    sid: int

    model_config = ConfigDict(protected_namespaces=())


class EffectedVocalsMetaData(BaseModel):
    """
    Metadata for an effected vocals track.

    Attributes
    ----------
    vocals_track : FileMetaData
        Metadata for the vocals track that effects were applied to.
    room_size : float
        The room size of the reverb effect applied to the vocals track.
    wet_level : float
        The wetness level of the reverb effect applied to the vocals
        track.
    dry_level : float
        The dryness level of the reverb effect. applied to the vocals
        track.
    damping : float
        The damping of the reverb effect applied to the vocals track.

    """

    vocals_track: FileMetaData
    room_size: float
    wet_level: float
    dry_level: float
    damping: float


class PitchShiftMetaData(BaseModel):
    """
    Metadata for a pitch-shifted audio track.

    Attributes
    ----------
    audio_track : FileMetaData
        Metadata for the audio track that was pitch-shifted.
    n_semitones : int
        The number of semitones the audio track was pitch-shifted by.

    """

    audio_track: FileMetaData
    n_semitones: int


class StagedAudioMetaData(BaseModel):
    """
    Metadata for a staged audio track.

    Attributes
    ----------
    audio_track : FileMetaData
        Metadata for the audio track that was staged.
    gain : float
        The gain applied to the audio track.

    """

    audio_track: FileMetaData
    gain: float


class MixedAudioMetaData(BaseModel):
    """
    Metadata for mixed audio.

    Attributes
    ----------
    staged_audio_tracks : list[StagedAudioMetaData]
        Metadata for the staged audio tracks that were mixed.

    output_sr : int
        The sample rate of the mixed audio.
    output_format : AudioExt
        The audio file format of the mixed audio.

    """

    staged_audio_tracks: list[StagedAudioMetaData]
    output_sr: int
    output_format: AudioExt


class EdgeTTSAudioMetaData(BaseModel):
    """
    Metadata for an audio track generated by Edge TTS.

    Attributes
    ----------
    text: str
        The text that was spoken to generate the audio track.
    file : FileMetaData, optional
        Metadata for file containing the text that was spoken to
        generate the audio track.
    voice : str
        The short name of the voice used for generating the audio track.
    pitch_shift : int
        The number of hertz the pitch of the voice speaking the
        provided text was shifted.
    speed_change : int
        The percentual change to the speed of the voice speaking the
        provided text.
    volume_change : int
        The percentual change to the volume of the voice speaking the
        provided text.

    """

    text: str
    file: FileMetaData | None
    voice: str
    pitch_shift: int
    speed_change: int
    volume_change: int


class MixedAudioType(StrEnum):
    """The valid types of mixed audio."""

    AUDIO = "audio"
    SONG = "song"
    SPEECH = "speech"


EdgeTTSVoiceTable = list[list[str]]
EdgeTTSVoiceKey = Literal[
    "Name",
    "ShortName",
    "DisplayName",
    "LocalName",
    "LocaleName",
    "Locale",
    "Gender",
    "WordsPerMinute",
    "Status",
]
EdgeTTSVoiceTagKey = Literal[
    "ContentCategories",
    "VoicePersonalities",
]
EdgeTTSKeys = list[EdgeTTSVoiceKey | EdgeTTSVoiceTagKey]



"""Module defining common component configurations for UI tabs."""

from __future__ import annotations

from pydantic import BaseModel

from ultimate_rvc.typing_extra import (
    AudioExt,
    AudioSplitMethod,
    EmbedderModel,
    F0Method,
    IndexAlgorithm,
    PretrainedType,
    SampleRate,
    TrainingF0Method,
    TrainingSampleRate,
    Vocoder,
)
from ultimate_rvc.web.config.component import (
    CheckboxConfig,
    DropdownConfig,
    NumberConfig,
    SliderConfig,
    TextboxConfig,
)
from ultimate_rvc.web.typing_extra import DatasetType, SongSourceType, SpeechSourceType


class BaseTabConfig(BaseModel):
    """
    Base model defining common component configuration settings for
    UI tabs.

    Attributes
    ----------
    embedder_model : DropdownConfig
        Configuration settings for an embedder model dropdown component.
    custom_embedder_model : DropdownConfig
        Configuration settings for a custom embedder model dropdown
        component.

    """

    embedder_model: DropdownConfig = DropdownConfig(
        label="Embedder model",
        info="The model to use for generating speaker embeddings.",
        value=EmbedderModel.CONTENTVEC,
        choices=list(EmbedderModel),
        exclude_value=True,
    )
    custom_embedder_model: DropdownConfig = DropdownConfig(
        label="Custom embedder model",
        info="Select a custom embedder model from the dropdown.",
        value=None,
        visible=False,
        render=False,
        exclude_value=True,
    )


class GenerationConfig(BaseTabConfig):
    """
    Common component configuration settings for generation tabs.

    voice_model : DropdownConfig
        Configuration settings for a voice model dropdown component.
    f0_methods : DropdownConfig
        Configuration settings for a pitch extraction algorithms
        dropdown component.
    index_rate : SliderConfig
        Configuration settings for an index rate slider component.
    rms_mix_rate : SliderConfig
        Configuration settings for a RMS mix rate slider component.
    protect_rate : SliderConfig
        Configuration settings for a protect rate slider component.
    split_voice : CheckboxConfig
        Configuration settings for a split voice checkbox component.
    autotune_voice: CheckboxConfig
        Configuration settings for an autotune voice checkbox component.
    autotune_strength: SliderConfig
        Configuration settings for an autotune strength slider
        component.
    sid : NumberConfig
        Configuration settings for a speaker ID number component.
    output_sr : DropdownConfig
        Configuration settings for an output sample rate dropdown
        component.
    output_format : DropdownConfig
        Configuration settings for an output format dropdown
        component.
    output_name : TextboxConfig
        Configuration settings for an output name textbox component.

    See Also
    --------
    BaseTabConfig
        Parent model defining common component configuration settings
        for UI tabs.

    """

    voice_model: DropdownConfig = DropdownConfig(
        label="Voice model",
        info="Select a model to use for voice conversion.",
        value=None,
        render=False,
        exclude_value=True,
    )
    f0_methods: DropdownConfig = DropdownConfig(
        label="Pitch extraction algorithm(s)",
        info=(
            "If more than one method is selected, then the median of the pitch values"
            " extracted by each method is used. RMVPE is recommended for most cases and"
            " is the default when no method is selected."
        ),
        value=[F0Method.RMVPE],
        choices=list(F0Method),
        multiselect=True,
    )
    index_rate: SliderConfig = SliderConfig(
        label="Index rate",
        info=(
            "Increase to bias the conversion towards the accent of the voice model."
            " Decrease to potentially reduce artifacts coming from the voice"
            " model.<br><br><br>"
        ),
        value=0.3,
        minimum=0.0,
        maximum=1.0,
    )
    rms_mix_rate: SliderConfig = SliderConfig(
        label="RMS mix rate",
        info=(
            "How much to mimic the loudness (0) of the input voice or a fixed loudness"
            " (1). A value of 1 is recommended for most cases.<br><br>"
        ),
        value=1.0,
        minimum=0.0,
        maximum=1.0,
    )
    protect_rate: SliderConfig = SliderConfig(
        label="Protect rate",
        info=(
            "Controls the extent to which consonants and breathing sounds are protected"
            " from artifacts. A higher value offers more protection but may worsen the"
            " indexing effect.<br><br>"
        ),
        value=0.33,
        minimum=0.0,
        maximum=0.5,
    )

    hop_length: SliderConfig = SliderConfig.hop_length(
        label="Hop length",
        info=(
            "How often the CREPE-based pitch extraction method checks for pitch changes"
            " measured in milliseconds. Lower values lead to longer conversion times"
            " and a higher risk of voice cracks, but better pitch accuracy."
        ),
        visible=True,
    )

    split_voice: CheckboxConfig = CheckboxConfig(
        label="Split input voice",
        info=(
            "Whether to split the input voice track into smaller segments before"
            " converting it. This can improve output quality for longer voice tracks."
        ),
        value=False,
    )
    autotune_voice: CheckboxConfig = CheckboxConfig(
        label="Autotune converted voice",
        info="Whether to apply autotune to the converted voice.<br><br>",
        value=False,
        exclude_value=True,
    )
    autotune_strength: SliderConfig = SliderConfig(
        label="Autotune intensity",
        info=(
            "Higher values result in stronger snapping to the chromatic grid and"
            " artifacting."
        ),
        value=1.0,
        minimum=0.0,
        maximum=1.0,
        visible=False,
    )
    sid: NumberConfig = NumberConfig(
        label="Speaker ID",
        info="Speaker ID for multi-speaker-models.",
        value=0,
        precision=0,
    )
    output_sr: DropdownConfig = DropdownConfig(
        label="Output sample rate",
        info="The sample rate of the mixed output track.",
        value=SampleRate.HZ_44100,
        choices=list(SampleRate),
    )
    output_format: DropdownConfig = DropdownConfig(
        label="Output format",
        info="The audio format of the mixed output track.",
        value=AudioExt.MP3,
        choices=list(AudioExt),
    )
    output_name: TextboxConfig = TextboxConfig(
        label="Output name",
        info="If no name is provided, a suitable name will be generated automatically.",
        value=None,
        placeholder="Ultimate RVC output",
        exclude_value=True,
    )


class SongGenerationConfig(GenerationConfig):
    """
    Common component configuration settings for song generation tabs.

    Attributes
    ----------
    source_type : DropdownConfig
        Configuration settings for a source type dropdown component.
    source : TextboxConfig
        Configuration settings for an input source textbox component.
    cached_song : DropdownConfig
        Configuration settings for a cached song dropdown component.
    clean_strength : SliderConfig
        Configuration settings for a clean strength slider component.
    clean_voice : CheckboxConfig
        Configuration settings for a clean voice checkbox component.
    room_size : SliderConfig
        Configuration settings for a room size slider component.
    wet_level : SliderConfig
        Configuration settings for a wetness level slider component.
    dry_level : SliderConfig
        Configuration settings for a dryness level slider component.
    damping : SliderConfig
        Configuration settings for a damping level slider component.
    main_gain : SliderConfig
        Configuration settings for a main gain slider component.
    inst_gain : SliderConfig
        Configuration settings for an instrumentals gain slider
        component.
    backup_gain : SliderConfig
        Configuration settings for a backup vocals gain slider
        component.

    See Also
    --------
    GenerationConfig
        Parent model defining common component configuration settings
        for song generation tabs.

    """

    source_type: DropdownConfig = DropdownConfig(
        label="Source type",
        info="The type of source to retrieve a song from.",
        value=SongSourceType.PATH,
        choices=list(SongSourceType),
        type="index",
        exclude_value=True,
    )
    source: TextboxConfig = TextboxConfig(
        label="Source",
        info="Link to a song on YouTube or the full path of a local audio file.",
        value=None,
        exclude_value=True,
    )
    cached_song: DropdownConfig = DropdownConfig(
        label="Source",
        info="Select a song from the list of cached songs.",
        value=None,
        visible=False,
        render=False,
        exclude_value=True,
    )
    clean_voice: CheckboxConfig = CheckboxConfig(
        label="Clean converted voice",
        info=(
            "Whether to clean the converted voice using noise reduction"
            " algorithms.<br><br>"
        ),
        value=False,
        exclude_value=True,
    )
    clean_strength: SliderConfig = SliderConfig.clean_strength(visible=False)
    room_size: SliderConfig = SliderConfig(
        label="Room size",
        info=(
            "Size of the room which reverb effect simulates. Increase for longer reverb"
            " time."
        ),
        value=0.15,
        minimum=0.0,
        maximum=1.0,
    )
    wet_level: SliderConfig = SliderConfig(
        label="Wetness level",
        info="Loudness of converted vocals with reverb effect applied.",
        value=0.2,
        minimum=0.0,
        maximum=1.0,
    )
    dry_level: SliderConfig = SliderConfig(
        label="Dryness level",
        info="Loudness of converted vocals without reverb effect applied.",
        value=0.8,
        minimum=0.0,
        maximum=1.0,
    )
    damping: SliderConfig = SliderConfig(
        label="Damping level",
        info="Absorption of high frequencies in reverb effect.",
        value=0.7,
        minimum=0.0,
        maximum=1.0,
    )
    main_gain: SliderConfig = SliderConfig.gain(
        label="Main gain",
        info="The gain to apply to the main vocals.",
    )
    inst_gain: SliderConfig = SliderConfig.gain(
        label="Instrumentals gain",
        info="The gain to apply to the instrumentals.",
    )
    backup_gain: SliderConfig = SliderConfig.gain(
        label="Backup gain",
        info="The gain to apply to the backup vocals.",
    )


class SpeechGenerationConfig(GenerationConfig):
    """
    Common component configuration settings for speech generation tabs.

    Attributes
    ----------
    source_type : DropdownConfig
        Configuration settings for a source type dropdown component.
    source : TextboxConfig
        Configuration settings for an input source textbox component.
    edge_tts_voice : DropdownConfig
        Configuration settings for an Edge TTS voice dropdown
        component.
    n_octaves : SliderConfig
        Configuration settings for an octave pitch shift slider
        component.
    n_semitones : SliderConfig
        Configuration settings for a semitone pitch shift slider
        component.
    tts_pitch_shift : SliderConfig
        Configuration settings for a TTS pitch shift slider
        component.
    tts_speed_change : SliderConfig
        Configuration settings for a TTS speed change slider
        component.
    tts_volume_change : SliderConfig
        Configuration settings for a TTS volume change slider
        component.
    clean_voice : CheckboxConfig
        Configuration settings for a clean voice checkbox
        component.
    clean_strength : SliderConfig
        Configuration settings for a clean strength slider
        component.
    output_gain : GainSliderConfig
        Configuration settings for an output gain slider component.

    See Also
    --------
    GenerationConfig
        Parent model defining common component configuration settings
        for generation tabs.

    """

    source_type: DropdownConfig = DropdownConfig(
        label="Source type",
        info="The type of source to generate speech from.",
        value=SpeechSourceType.TEXT,
        choices=list(SpeechSourceType),
        type="index",
        exclude_value=True,
    )
    source: TextboxConfig = TextboxConfig(
        label="Source",
        info="Text to generate speech from",
        value=None,
        exclude_value=True,
    )
    edge_tts_voice: DropdownConfig = DropdownConfig(
        label="Edge TTS voice",
        info="Select a voice to use for text to speech conversion.",
        value=None,
        render=False,
        exclude_value=True,
    )
    n_octaves: SliderConfig = SliderConfig.octave_shift(
        label="Octave shift",
        info=(
            "The number of octaves to pitch-shift the converted speech by. Use 1 for"
            " male-to-female and -1 for vice-versa."
        ),
    )
    n_semitones: SliderConfig = SliderConfig.semitone_shift(
        label="Semitone shift",
        info="The number of semi-tones to pitch-shift the converted speech by.",
    )
    tts_pitch_shift: SliderConfig = SliderConfig(
        label="Edge TTS pitch shift",
        info=(
            "The number of hertz to shift the pitch of the speech generated by Edge"
            " TTS."
        ),
        value=0,
        minimum=-100,
        maximum=100,
        step=1,
    )
    tts_speed_change: SliderConfig = SliderConfig(
        label="TTS speed change",
        info="The percentual change to the speed of the speech generated by Edge TTS.",
        value=0,
        minimum=-50,
        maximum=100,
        step=1,
    )
    tts_volume_change: SliderConfig = SliderConfig(
        label="TTS volume change",
        info="The percentual change to the volume of the speech generated by Edge TTS.",
        value=0,
        minimum=-100,
        maximum=100,
        step=1,
    )
    clean_voice: CheckboxConfig = CheckboxConfig(
        label="Clean converted voice",
        info=(
            "Whether to clean the converted voice using noise reduction"
            " algorithms.<br><br>"
        ),
        value=True,
        exclude_value=True,
    )
    clean_strength: SliderConfig = SliderConfig.clean_strength(visible=True)
    output_gain: SliderConfig = SliderConfig.gain(
        label="Output gain",
        info="The gain to apply to the converted speech.<br><br>",
    )


class TrainingConfig(BaseTabConfig):
    """
    Common component configuration settings for training tabs.

    Attributes
    ----------
    dataset_type : DropdownConfig
        Configuration settings for a dataset type dropdown component.
    dataset : DropdownConfig
        Configuration settings for a dataset dropdown component.
    dataset_name : TextboxConfig
        Configuration settings for a dataset name textbox component.
    preprocess_model : DropdownConfig
        Configuration settings for a model name dropdown component
        for audio preprocessing.
    sample_rate : DropdownConfig
        Configuration settings for a sample rate dropdown component.
    filter_audio : CheckboxConfig
        Configuration settings for a filter audio checkbox component.
    clean_audio : CheckboxConfig
        Configuration settings for a clean audio checkbox component.
    clean_strength : SliderConfig
        Configuration settings for a clean strength slider component.
    split_method : DropdownConfig
        Configuration settings for an audio splitting method dropdown
        component.
    chunk_len : SliderConfig
        Configuration settings for a chunk length slider component.
    overlap_len : SliderConfig
        Configuration settings for an overlap length slider component.
    preprocess_cores : SliderConfig
        Configuration settings for a CPU cores slider component for
        preprocessing.
    extract_model : DropdownConfig
        Configuration settings for a model name dropdown component for
        feature extraction.
    f0_method : DropdownConfig
        Configuration settings for an F0 method dropdown component.
    hop_length : SliderConfig
        Configuration settings for a hop length slider component.
    include_mutes : SliderConfig
        Configuration settings for an include mutes slider component.
    extract_cores : SliderConfig
        Configuration settings for a CPU cores slider component for
        feature extraction.
    extraction_acceleration : HardwareAccelerationConfig
        Configuration settings for a hardware acceleration component for
        feature extraction.
    extraction_gpus : DropdownConfig
        Configuration settings for a GPU dropdown compoennt for feature
        extraction.
    train_model : DropdownConfig
        Configuration settings for a model name dropdown component for
        training.
    num_epochs : SliderConfig
        Configuration settings for a number of epochs slider component.
    batch_size : SliderConfig
        Configuration settings for a batch size slider component.
    detect_overtraining : CheckboxConfig
        Configuration settings for a detect overtraining checkbox
        component.
    overtraining_threshold : SliderConfig
        Configuration settings for an overtraining threshold slider
        component.
    vocoder : DropdownConfig
        Configuration settings for a vocoder dropdown component.
    index_algorithm : DropdownConfig
        Configuration settings for an index algorithm dropdown
        component.
    pretrained_type : DropdownConfig
        Configuration settings for a pretrained model type dropdown
        component.
    custom_pretrained_model : DropdownConfig
        Configuration settings for a custom pretrained model dropdown
        component.
    save_interval : SliderConfig
        Configuration settings for a save-interval slider component.
    save_all_checkpoints : CheckboxConfig
        Configuration settings for a save-all-checkpoints checkbox
        component.
    save_all_weights : CheckboxConfig
        Configuration settings for a save-all-weights checkbox
        component.
    clear_saved_data : CheckboxConfig
        Configuration settings for a clear-saved-data checkbox
        component.
    upload_model : CheckboxConfig
        Configuration settings for an upload voice model checkbox
        component.
    upload_name : TextboxConfig
        Configuration settings for an upload name textbox component.
    training_acceleration : HardwareAccelerationConfig
        Configuration settings for a hardware acceleration component for
        training.
    training_gpus : DropdownConfig
        Configuration settings for a GPU dropdown component for
        training.
    preload_dataset : CheckboxConfig
        Configuration settings for a preload dataset checkbox component.
    reduce_memory_usage : CheckboxConfig
        Configuration settings for a reduce-memory-usage checkbox
        component.

    See Also
    --------
    BaseTabConfig
        Parent model defining common component configuration settings
        for UI tabs.

    """

    dataset_type: DropdownConfig = DropdownConfig(
        label="Dataset type",
        info="Select the type of dataset to preprocess.",
        value=DatasetType.NEW_DATASET,
        choices=list(DatasetType),
        exclude_value=True,
    )
    dataset: DropdownConfig = DropdownConfig(
        label="Dataset path",
        info=(
            "The path to an existing dataset. Either select a path to a previously"
            " created dataset or provide a path to an external dataset."
        ),
        value=None,
        allow_custom_value=True,
        visible=False,
        render=False,
        exclude_value=True,
    )
    dataset_name: TextboxConfig = TextboxConfig(
        label="Dataset name",
        info=(
            "The name of the new dataset. If the dataset already exists, the provided"
            " audio files will be added to it."
        ),
        value="My dataset",
        exclude_value=True,
    )
    preprocess_model: DropdownConfig = DropdownConfig(
        label="Model name",
        info=(
            "Name of the model to preprocess the given dataset for. Either select an"
            " existing model from the dropdown or provide the name of a new model."
        ),
        value="My model",
        allow_custom_value=True,
        render=False,
        exclude_value=True,
    )
    sample_rate: DropdownConfig = DropdownConfig(
        label="Sample rate",
        info="Target sample rate for the audio files in the provided dataset.",
        value=TrainingSampleRate.HZ_40K,
        choices=list(TrainingSampleRate),
    )
    filter_audio: CheckboxConfig = CheckboxConfig(
        label="Filter audio",
        info=(
            "Whether to remove low-frequency sounds from the audio files in the"
            " provided dataset by applying a high-pass butterworth filter.<br><br>"
        ),
        value=True,
    )
    clean_audio: CheckboxConfig = CheckboxConfig(
        label="Clean audio",
        info=(
            "Whether to clean the audio files in the provided dataset using noise"
            " reduction algorithms.<br><br><br>"
        ),
        value=False,
        exclude_value=True,
    )
    clean_strength: SliderConfig = SliderConfig.clean_strength(visible=False)
    split_method: DropdownConfig = DropdownConfig(
        label="Audio splitting method",
        info=(
            "The method to use for splitting the audio files in the provided dataset."
            " Use the `Skip` method to skip splitting if the audio files are already"
            " split. Use the `Simple` method if excessive silence has already been"
            " removed from the audio files. Use the `Automatic` method for automatic"
            " silence detection and splitting around it."
        ),
        value=AudioSplitMethod.AUTOMATIC,
        choices=list(AudioSplitMethod),
        exclude_value=True,
    )
    chunk_len: SliderConfig = SliderConfig(
        label="Chunk length",
        info="Length of split audio chunks.",
        value=3.0,
        minimum=0.5,
        maximum=5.0,
        step=0.1,
        visible=False,
    )
    overlap_len: SliderConfig = SliderConfig(
        label="Overlap length",
        info="Length of overlap between split audio chunks.",
        value=0.3,
        minimum=0.0,
        maximum=0.4,
        step=0.1,
        visible=False,
    )
    preprocess_cores: SliderConfig = SliderConfig.cpu_cores()

    extract_model: DropdownConfig = DropdownConfig(
        label="Model name",
        info=(
            "Name of the model with an associated preprocessed dataset to extract"
            " training features from. When a new dataset is preprocessed, its"
            " associated model is selected by default."
        ),
        value=None,
        render=False,
        exclude_value=True,
    )
    f0_method: DropdownConfig = DropdownConfig(
        label="F0 method",
        info="The method to use for extracting pitch features.",
        value=TrainingF0Method.RMVPE,
        choices=list(TrainingF0Method),
        exclude_value=True,
    )

    hop_length: SliderConfig = SliderConfig.hop_length(
        label="Hop length",
        info="The hop length to use for extracting pitch features.<br><br>",
        visible=False,
    )
    include_mutes: SliderConfig = SliderConfig(
        label="Include mutes",
        info=(
            "The number of mute audio files to include in the generated training file"
            " list. Adding silent files enables the training model to handle pure"
            " silence in inferred audio files. If the preprocessed audio dataset"
            " already contains segments of pure silence, set this to 0."
        ),
        value=2,
        minimum=0,
        maximum=10,
        step=1,
    )
    extraction_cores: SliderConfig = SliderConfig.cpu_cores()
    extraction_acceleration: DropdownConfig = DropdownConfig.hardware_acceleration()
    extraction_gpus: DropdownConfig = DropdownConfig.gpu()

    train_model: DropdownConfig = DropdownConfig(
        label="Model name",
        info=(
            "Name of the model to train. When training features are extracted for a new"
            " model, its name is selected by default."
        ),
        value=None,
        render=False,
        exclude_value=True,
    )
    num_epochs: SliderConfig = SliderConfig(
        label="Number of epochs",
        info=(
            "The number of epochs to train the voice model. A higher number can improve"
            " voice model performance but may lead to overtraining."
        ),
        value=500,
        minimum=1,
        maximum=1000,
        step=1,
    )
    batch_size: SliderConfig = SliderConfig(
        label="Batch size",
        info=(
            "The number of samples in each training batch. It is advisable to align"
            " this value with the available VRAM of your GPU."
        ),
        value=8,
        minimum=1,
        maximum=64,
        step=1,
    )
    detect_overtraining: CheckboxConfig = CheckboxConfig(
        label="Detect overtraining",
        info=(
            "Whether to detect overtraining to prevent the voice model from learning"
            " the training data too well and losing the ability to generalize to new"
            " data."
        ),
        value=False,
        exclude_value=True,
    )
    overtraining_threshold: SliderConfig = SliderConfig(
        label="Overtraining threshold",
        info=(
            "The maximum number of epochs to continue training without any observed"
            " improvement in voice model performance."
        ),
        value=50,
        minimum=1,
        maximum=100,
        visible=False,
    )
    vocoder: DropdownConfig = DropdownConfig(
        label="Vocoder",
        info=(
            "The vocoder to use for audio synthesis during training. HiFi-GAN provides"
            " basic audio fidelity, while RefineGAN provides the highest audio"
            " fidelity."
        ),
        value=Vocoder.HIFI_GAN,
        choices=list(Vocoder),
    )
    index_algorithm: DropdownConfig = DropdownConfig(
        label="Index algorithm",
        info=(
            "The method to use for generating an index file for the trained voice"
            " model. `KMeans` is particularly useful for large datasets."
        ),
        value=IndexAlgorithm.AUTO,
        choices=list(IndexAlgorithm),
    )
    pretrained_type: DropdownConfig = DropdownConfig(
        label="Pretrained model type",
        info=(
            "The type of pretrained model to finetune the voice model on. `None` will"
            " train the voice model from scratch, while `Default` will use a pretrained"
            " model tailored to the specific voice model architecture. `Custom` will"
            " use a custom pretrained that you provide."
        ),
        value=PretrainedType.DEFAULT,
        choices=list(PretrainedType),
        exclude_value=True,
    )
    custom_pretrained_model: DropdownConfig = DropdownConfig(
        label="Custom pretrained model",
        info="Select a custom pretrained model to finetune from the dropdown.",
        value=None,
        visible=False,
        render=False,
        exclude_value=True,
    )
    save_interval: SliderConfig = SliderConfig(
        label="Save interval",
        info=(
            "The epoch interval at which to to save voice model weights and"
            " checkpoints. The best model weights are always saved regardless of this"
            " setting."
        ),
        value=10,
        minimum=1,
        maximum=100,
        step=1,
    )
    save_all_checkpoints: CheckboxConfig = CheckboxConfig(
        label="Save all checkpoints",
        info=(
            "Whether to save a unique checkpoint at each save interval. If not enabled,"
            " only the latest checkpoint will be saved at each interval."
        ),
        value=False,
    )
    save_all_weights: CheckboxConfig = CheckboxConfig(
        label="Save all weights",
        info=(
            "Whether to save unique voice model weights at each save interval. If not"
            " enabled, only the best voice model weights will be saved."
        ),
        value=False,
    )
    clear_saved_data: CheckboxConfig = CheckboxConfig(
        label="Clear saved data",
        info=(
            "Whether to delete any existing training data associated with the voice"
            " model before training commences. Enable this setting only if you are"
            " training a new voice model from scratch or restarting training."
        ),
        value=False,
    )
    upload_model: CheckboxConfig = CheckboxConfig(
        label="Upload voice model",
        info=(
            "Whether to automatically upload the trained voice model so that it can be"
            " used for generation tasks within the Ultimate RVC app."
        ),
        value=False,
        exclude_value=True,
    )
    upload_name: TextboxConfig = TextboxConfig(
        label="Upload name",
        info="The name to give the uploaded voice model.",
        value=None,
        visible=False,
        exclude_value=True,
    )
    training_acceleration: DropdownConfig = DropdownConfig.hardware_acceleration()
    training_gpus: DropdownConfig = DropdownConfig.gpu()
    preload_dataset: CheckboxConfig = CheckboxConfig(
        label="Preload dataset",
        info=(
            "Whether to preload all training data into GPU memory. This can improve"
            " training speed but requires a lot of VRAM.<br><br>"
        ),
        value=False,
    )
    reduce_memory_usage: CheckboxConfig = CheckboxConfig(
        label="Reduce memory usage",
        info=(
            "Whether to reduce VRAM usage at the cost of slower training speed by"
            " enabling activation checkpointing. This is useful for GPUs with limited"
            " memory (e.g., <6GB VRAM) or when training with a batch size larger than"
            " what your GPU can normally accommodate."
        ),
        value=False,
    )



import torch
from torch.nn.utils.parametrizations import spectral_norm, weight_norm
from torch.utils.checkpoint import checkpoint

from ultimate_rvc.rvc.lib.algorithm.commons import get_padding
from ultimate_rvc.rvc.lib.algorithm.residuals import LRELU_SLOPE


class MultiPeriodDiscriminator(torch.nn.Module):
    """
    Multi-period discriminator.

    This class implements a multi-period discriminator, which is used to
    discriminate between real and fake audio signals. The discriminator
    is composed of a series of convolutional layers that are applied to
    the input signal at different periods.

    Args:
        use_spectral_norm (bool): Whether to use spectral normalization.
            Defaults to False.

    """

    def __init__(self, use_spectral_norm: bool = False, checkpointing: bool = False):
        super().__init__()
        periods = [2, 3, 5, 7, 11, 17, 23, 37]
        self.checkpointing = checkpointing
        self.discriminators = torch.nn.ModuleList(
            [DiscriminatorS(use_spectral_norm=use_spectral_norm)]
            + [DiscriminatorP(p, use_spectral_norm=use_spectral_norm) for p in periods],
        )

    def forward(self, y, y_hat):
        y_d_rs, y_d_gs, fmap_rs, fmap_gs = [], [], [], []
        for d in self.discriminators:
            if self.training and self.checkpointing:
                y_d_r, fmap_r = checkpoint(d, y, use_reentrant=False)
                y_d_g, fmap_g = checkpoint(d, y_hat, use_reentrant=False)
            else:
                y_d_r, fmap_r = d(y)
                y_d_g, fmap_g = d(y_hat)
            y_d_rs.append(y_d_r)
            y_d_gs.append(y_d_g)
            fmap_rs.append(fmap_r)
            fmap_gs.append(fmap_g)

        return y_d_rs, y_d_gs, fmap_rs, fmap_gs


class DiscriminatorS(torch.nn.Module):
    """
    Discriminator for the short-term component.

    This class implements a discriminator for the short-term component
    of the audio signal. The discriminator is composed of a series of
    convolutional layers that are applied to the input signal.
    """

    def __init__(self, use_spectral_norm: bool = False):
        super().__init__()

        norm_f = spectral_norm if use_spectral_norm else weight_norm
        self.convs = torch.nn.ModuleList(
            [
                norm_f(torch.nn.Conv1d(1, 16, 15, 1, padding=7)),
                norm_f(torch.nn.Conv1d(16, 64, 41, 4, groups=4, padding=20)),
                norm_f(torch.nn.Conv1d(64, 256, 41, 4, groups=16, padding=20)),
                norm_f(torch.nn.Conv1d(256, 1024, 41, 4, groups=64, padding=20)),
                norm_f(torch.nn.Conv1d(1024, 1024, 41, 4, groups=256, padding=20)),
                norm_f(torch.nn.Conv1d(1024, 1024, 5, 1, padding=2)),
            ],
        )
        self.conv_post = norm_f(torch.nn.Conv1d(1024, 1, 3, 1, padding=1))
        self.lrelu = torch.nn.LeakyReLU(LRELU_SLOPE)

    def forward(self, x):
        fmap = []
        for conv in self.convs:
            x = self.lrelu(conv(x))
            fmap.append(x)
        x = self.conv_post(x)
        fmap.append(x)
        x = torch.flatten(x, 1, -1)
        return x, fmap


class DiscriminatorP(torch.nn.Module):
    """
    Discriminator for the long-term component.

    This class implements a discriminator for the long-term component
    of the audio signal. The discriminator is composed of a series of
    convolutional layers that are applied to the input signal at a given
    period.

    Args:
        period (int): Period of the discriminator.
        kernel_size (int): Kernel size of the convolutional layers. Defaults to 5.
        stride (int): Stride of the convolutional layers. Defaults to 3.
        use_spectral_norm (bool): Whether to use spectral normalization. Defaults to False.

    """

    def __init__(
        self,
        period: int,
        kernel_size: int = 5,
        stride: int = 3,
        use_spectral_norm: bool = False,
    ):
        super().__init__()
        self.period = period
        norm_f = spectral_norm if use_spectral_norm else weight_norm

        in_channels = [1, 32, 128, 512, 1024]
        out_channels = [32, 128, 512, 1024, 1024]

        self.convs = torch.nn.ModuleList(
            [
                norm_f(
                    torch.nn.Conv2d(
                        in_ch,
                        out_ch,
                        (kernel_size, 1),
                        (stride, 1),
                        padding=(get_padding(kernel_size, 1), 0),
                    ),
                )
                for in_ch, out_ch in zip(in_channels, out_channels, strict=False)
            ],
        )

        self.conv_post = norm_f(torch.nn.Conv2d(1024, 1, (3, 1), 1, padding=(1, 0)))
        self.lrelu = torch.nn.LeakyReLU(LRELU_SLOPE)

    def forward(self, x):
        fmap = []
        b, c, t = x.shape
        if t % self.period != 0:
            n_pad = self.period - (t % self.period)
            x = torch.nn.functional.pad(x, (0, n_pad), "reflect")
        x = x.view(b, c, -1, self.period)

        for conv in self.convs:
            x = self.lrelu(conv(x))
            fmap.append(x)
        x = self.conv_post(x)
        fmap.append(x)
        x = torch.flatten(x, 1, -1)
        return x, fmap



"""Module which defines functions to manage voice models."""

from __future__ import annotations

from typing import TYPE_CHECKING

import lazy_loader as lazy

import re
import shutil
import urllib.request
import zipfile
from pathlib import Path

from ultimate_rvc.common import (
    CUSTOM_EMBEDDER_MODELS_DIR,
    CUSTOM_PRETRAINED_MODELS_DIR,
    PRETRAINED_MODELS_DIR,
    TRAINING_MODELS_DIR,
    VOICE_MODELS_DIR,
)
from ultimate_rvc.core.common import (
    copy_files_to_new_dir,
    display_progress,
    get_file_size,
    json_dump,
    json_load,
    validate_model,
    validate_url,
)
from ultimate_rvc.core.exceptions import (
    Entity,
    Location,
    ModelEntity,
    ModelNotFoundError,
    NotFoundError,
    NotProvidedError,
    PretrainedModelExistsError,
    PretrainedModelNotAvailableError,
    UIMessage,
    UploadLimitError,
    UploadTypeError,
)
from ultimate_rvc.core.manage.common import delete_directory, get_items
from ultimate_rvc.core.manage.typing_extra import (
    PretrainedModelMetaDataTable,
    VoiceModelMetaData,
    VoiceModelMetaDataList,
    VoiceModelMetaDataPredicate,
    VoiceModelMetaDataTable,
    VoiceModelTagName,
)

if TYPE_CHECKING:
    from typing import NoReturn

    from _collections_abc import Sequence

    import requests

    import tqdm

    from ultimate_rvc.typing_extra import PretrainedSampleRate, StrPath
else:
    requests = lazy.load("requests")
    tqdm = lazy.load("tqdm")

PUBLIC_MODELS_JSON = json_load(Path(__file__).parent / "public_models.json")
PUBLIC_MODELS_TABLE = VoiceModelMetaDataTable.model_validate(PUBLIC_MODELS_JSON)


def get_voice_model_names() -> list[str]:
    """
    Get the names of all saved voice models.

    Returns
    -------
    list[str]
        A list of names of all saved voice models.

    """
    return get_items(VOICE_MODELS_DIR)


def get_custom_embedder_model_names() -> list[str]:
    """
    Get the names of all saved custom embedder models.

    Returns
    -------
    list[str]
        A list of the names of all saved custom embedder models.

    """
    return get_items(CUSTOM_EMBEDDER_MODELS_DIR)


def get_custom_pretrained_model_names() -> list[str]:
    """
    Get the names of all saved custom pretrained models.

    Returns
    -------
    list[str]
        A list of the names of all saved custom pretrained models.

    """
    return get_items(CUSTOM_PRETRAINED_MODELS_DIR)


def get_training_model_names() -> list[str]:
    """
    Get the names of all saved training models.

    Returns
    -------
    list[str]
        A list of the names of all saved training models.

    """
    return get_items(TRAINING_MODELS_DIR)


def load_public_models_table(
    predicates: Sequence[VoiceModelMetaDataPredicate],
) -> VoiceModelMetaDataList:
    """
    Load table containing metadata of public voice models, optionally
    filtered by a set of predicates.

    Parameters
    ----------
    predicates : Sequence[VoiceModelMetaDataPredicate]
        Predicates to filter the metadata table by.

    Returns
    -------
    VoiceModelMetaDataList
        List containing metadata for each public voice model that
        satisfies the given predicates.

    """
    return [
        [
            model.name,
            model.description,
            model.tags,
            model.credit,
            model.added,
            model.url,
        ]
        for model in PUBLIC_MODELS_TABLE.models
        if all(predicate(model) for predicate in predicates)
    ]


def get_public_model_tags() -> list[VoiceModelTagName]:
    """
    Get the names of all valid public voice model tags.

    Returns
    -------
    list[str]
        A list of names of all valid public voice model tags.

    """
    return [tag.name for tag in PUBLIC_MODELS_TABLE.tags]


def filter_public_models_table(
    tags: Sequence[str],
    query: str,
) -> VoiceModelMetaDataList:
    """
    Filter table containing metadata of public voice models by tags and
    a search query.


    The search query is matched against the name, description, tags,
    credit,and added date of each entry in the metadata table. Case
    insensitive search is performed. If the search query is empty, the
    metadata table is filtered only bythe given tags.

    Parameters
    ----------
    tags : Sequence[str]
        Tags to filter the metadata table by.
    query : str
        Search query to filter the metadata table by.

    Returns
    -------
    VoiceModelMetaDataList
        List containing metadata for each public voice model that
        match the given tags and search query.

    """

    def _tags_predicate(model: VoiceModelMetaData) -> bool:
        return all(tag in model.tags for tag in tags)

    def _query_predicate(model: VoiceModelMetaData) -> bool:
        return (
            query.lower()
            in (
                f"{model.name} {model.description} {' '.join(model.tags)} "
                f"{model.credit} {model.added}"
            ).lower()
            if query
            else True
        )

    filter_fns = [_tags_predicate, _query_predicate]

    return load_public_models_table(filter_fns)


def get_pretrained_metadata() -> PretrainedModelMetaDataTable:
    """
    Get metadata for pretrained models available for download.

    Returns
    -------
    PretrainedModelMetaDataTable
        Table with metadata for pretrained models available for
        download.

    """
    PRETRAINED_MODELS_DIR.mkdir(parents=True, exist_ok=True)
    pretrained_metadata_path = PRETRAINED_MODELS_DIR / "custom_pretrains.json"
    if pretrained_metadata_path.is_file():
        pretrained_metadata_dict = json_load(pretrained_metadata_path)
        pretrained_metadata = PretrainedModelMetaDataTable.model_validate(
            pretrained_metadata_dict,
        )
    else:
        try:
            json_url = "https://huggingface.co/JackismyShephard/ultimate-rvc/raw/main/pretrains.json"
            response = requests.get(json_url, timeout=10)
            response.raise_for_status()
            pretrained_metadata_dict = response.json()
            pretrained_metadata = PretrainedModelMetaDataTable.model_validate(
                pretrained_metadata_dict,
            )
            json_dump(pretrained_metadata_dict, pretrained_metadata_path)
        except requests.exceptions.RequestException:
            pretrained_metadata_dict = {
                "Titan": {
                    "32k": {
                        "D": (
                            "blaise-tk/TITAN/resolve/main/models/medium/"
                            "32k/pretrained/D-f032k-TITAN-Medium.pth"
                        ),
                        "G": (
                            "blaise-tk/TITAN/resolve/main/models/medium/"
                            "32k/pretrained/G-f032k-TITAN-Medium.pth"
                        ),
                    },
                },
            }
            pretrained_metadata = PretrainedModelMetaDataTable.model_validate(
                pretrained_metadata_dict,
            )
    return pretrained_metadata


def _extract_voice_model(
    zip_file: StrPath,
    extraction_dir: StrPath,
    remove_incomplete: bool = True,
    remove_zip: bool = False,
) -> None:
    """
    Extract a zipped voice model to a directory.

    Parameters
    ----------
    zip_file : StrPath
        The path to a zip file containing the voice model to extract.
    extraction_dir : StrPath
        The path to the directory to extract the voice model to.

    remove_incomplete : bool, default=True
        Whether to remove the extraction directory if the extraction
        process fails.
    remove_zip : bool, default=False
        Whether to remove the zip file once the extraction process is
        complete.

    Raises
    ------
    NotFoundError
        If no model file is found in the extracted zip file.

    """
    extraction_path = Path(extraction_dir)
    zip_path = Path(zip_file)
    extraction_completed = False
    try:
        extraction_path.mkdir(parents=True)
        with zipfile.ZipFile(zip_path, "r") as zip_ref:
            zip_ref.extractall(extraction_path)
        file_path_map = {
            ext: Path(root, name)
            for root, _, files in extraction_path.walk()
            for name in files
            for ext in [".index", ".pth"]
            if Path(name).suffix == ext
            and Path(root, name).stat().st_size
            > 1024 * (100 if ext == ".index" else 1024 * 40)
        }
        if ".pth" not in file_path_map:
            raise NotFoundError(
                entity=Entity.MODEL_FILE,
                location=Location.EXTRACTED_ZIP_FILE,
                is_path=False,
            )

        # move model and index file to root of the extraction directory
        for file_path in file_path_map.values():
            file_path.rename(extraction_path / file_path.name)

        # remove any sub-directories within the extraction directory
        for path in extraction_path.iterdir():
            if path.is_dir():
                shutil.rmtree(path)
        extraction_completed = True
    finally:
        if not extraction_completed and remove_incomplete and extraction_path.is_dir():
            shutil.rmtree(extraction_path)
        if remove_zip and zip_path.exists():
            zip_path.unlink()


def download_voice_model(url: str, name: str) -> None:
    """
    Download a zipped voice model.

    Parameters
    ----------
    url : str
        An URL pointing to a location where the zipped voice model can
        be downloaded from.
    name : str
        The name to give to the downloaded voice model.

    """
    validate_url(url)
    extraction_path = validate_model(name, Entity.VOICE_MODEL, mode="not_exists")

    zip_name = url.rsplit("/", maxsplit=1)[-1].split("?", maxsplit=1)[0]

    # NOTE in case huggingface link is a direct link rather
    # than a resolve link then convert it to a resolve link
    url = re.sub(
        r"https://huggingface.co/([^/]+)/([^/]+)/blob/(.*)",
        r"https://huggingface.co/\1/\2/resolve/\3",
        url,
    )
    if "pixeldrain.com" in url:
        url = f"https://pixeldrain.com/api/file/{zip_name}"

    display_progress("[~] Downloading voice model ...")
    urllib.request.urlretrieve(url, zip_name)  # noqa: S310

    display_progress("[~] Extracting zip file...")
    _extract_voice_model(zip_name, extraction_path, remove_zip=True)


def _download_pretrained_model_file(
    url: str,
    destination: StrPath,
    progress_bar: tqdm.tqdm[NoReturn] | None = None,
) -> None:
    """
    Download a pretrained model file.

    Parameters
    ----------
    url : str
        The URL of the pretrained model file to download.
    destination : strPath
        The destination to save the downloaded pretrained model file to.
    progress_bar : tqdm.tqdm, optional
        TQDM progress bar to update.

    """
    response = requests.get(url, stream=True, timeout=10)
    response.raise_for_status()
    block_size = 1024
    destination_path = Path(destination)
    with destination_path.open("wb") as file:
        for data in response.iter_content(block_size):
            file.write(data)
            if progress_bar:
                progress_bar.update(len(data))


def download_pretrained_model(name: str, sample_rate: PretrainedSampleRate) -> None:
    """
    Download a pretrained model.

    Parameters
    ----------
    name : str
        The name of the pretrained model to download.
    sample_rate : PretrainedSampleRate
        The sample rate of the pretrained model to download.

    Raises
    ------
    PretrainedModelExistsError
        If a pretrained model with the provided name and sample rate
        already exists.
    PretrainedModelNotAvailableError
        If a pretrained model with the provided name and sample rate is
        not available for download.

    """
    CUSTOM_PRETRAINED_MODELS_DIR.mkdir(parents=True, exist_ok=True)
    model_path = CUSTOM_PRETRAINED_MODELS_DIR / f"{name} {sample_rate}"
    if model_path.is_dir():
        raise PretrainedModelExistsError(name, sample_rate)

    pretrained_metadata = get_pretrained_metadata()

    if name not in pretrained_metadata.names:
        raise PretrainedModelNotAvailableError(name)
    if sample_rate not in pretrained_metadata.get_sample_rates(name):
        raise PretrainedModelNotAvailableError(name, sample_rate)
    paths = pretrained_metadata[name][sample_rate]

    d_url = f"https://huggingface.co/{paths.D}"
    g_url = f"https://huggingface.co/{paths.G}"

    total_size = get_file_size(d_url) + get_file_size(g_url)

    model_path.mkdir(parents=True)
    import concurrent.futures  # noqa: PLC0415

    with (
        tqdm.tqdm(
            total=total_size,
            unit="iB",
            unit_scale=True,
            desc="Downloading files",
        ) as tqdm_bar,
        concurrent.futures.ThreadPoolExecutor() as executor,
    ):
        futures = [
            executor.submit(
                _download_pretrained_model_file,
                d_url,
                model_path / Path(paths.D).name,
                tqdm_bar,
            ),
            executor.submit(
                _download_pretrained_model_file,
                g_url,
                model_path / Path(paths.G).name,
                tqdm_bar,
            ),
        ]

        for future in futures:
            try:
                future.result()
            except requests.exceptions.RequestException as e:
                shutil.rmtree(model_path)
                raise PretrainedModelNotAvailableError(name, sample_rate) from e


def upload_voice_model(files: Sequence[StrPath], name: str) -> None:
    """
    Upload a voice model from either a zip file or a .pth file and an
    optional index file.

    Parameters
    ----------
    files : Sequence[StrPath]
        Paths to the files to upload.
    name : str
        The name to give to the uploaded voice model.

    Raises
    ------
    NotProvidedError
        If no file paths or name are provided.
    UploadTypeError
        If a single uploaded file is not a .pth file or a .zip file.
        If two uploaded files are not a .pth file and an .index file.
    UploadLimitError
        If more than two file paths are provided.

    """
    if not files:
        raise NotProvidedError(entity=Entity.FILES, ui_msg=UIMessage.NO_UPLOADED_FILES)
    model_dir_path = validate_model(name, Entity.VOICE_MODEL, mode="not_exists")
    sorted_file_paths = sorted([Path(f) for f in files], key=lambda f: f.suffix)
    match sorted_file_paths:
        case [file_path]:
            if file_path.suffix == ".pth":
                copy_files_to_new_dir([file_path], model_dir_path)
            # NOTE a .pth file is actually itself a zip file
            elif zipfile.is_zipfile(file_path):
                _extract_voice_model(file_path, model_dir_path)
            else:
                raise UploadTypeError(
                    entity=Entity.FILES,
                    valid_types=[".pth", ".zip"],
                    type_class="formats",
                    multiple=False,
                )
        case [index_path, pth_path]:
            if index_path.suffix == ".index" and pth_path.suffix == ".pth":
                copy_files_to_new_dir([index_path, pth_path], model_dir_path)
            else:
                raise UploadTypeError(
                    entity=Entity.FILES,
                    valid_types=[".pth", ".index"],
                    type_class="formats",
                    multiple=True,
                )
        case _:
            raise UploadLimitError(entity=Entity.FILES, limit="two")


def _extract_custom_embedder_model(
    zip_file: StrPath,
    extraction_dir: StrPath,
    remove_incomplete: bool = True,
    remove_zip: bool = False,
) -> None:
    """
    Extract a zipped custom embedder model to a directory.

    Parameters
    ----------
    zip_file : StrPath
        The path to a zip file containing the custom embedder model to
        extract.
    extraction_dir : StrPath
        The path to the directory to extract the custom embedder model
        to.

    remove_incomplete : bool, default=True
        Whether to remove the extraction directory if the extraction
        process fails.
    remove_zip : bool, default=False
        Whether to remove the zip file once the extraction process is
        complete.

    Raises
    ------
    NotFoundError
        If no pytorch_model.bin file or config.json file is found in
        the extracted zip file.

    """
    extraction_path = Path(extraction_dir)
    zip_path = Path(zip_file)
    extraction_completed = False
    try:
        extraction_path.mkdir(parents=True)
        with zipfile.ZipFile(zip_path, "r") as zip_ref:
            zip_ref.extractall(extraction_path)
        file_path_map = {
            file: Path(root, file)
            for root, _, files in extraction_path.walk()
            for file in files
            if file in {"pytorch_model.bin", "config.json"}
        }
        if "config.json" not in file_path_map:
            raise NotFoundError(
                entity=Entity.CONFIG_JSON_FILE,
                location=Location.EXTRACTED_ZIP_FILE,
                is_path=False,
            )
        if "pytorch_model.bin" not in file_path_map:
            raise NotFoundError(
                entity=Entity.MODEL_BIN_FILE,
                location=Location.EXTRACTED_ZIP_FILE,
                is_path=False,
            )

        # move pytorch_model.bin file and config.json file to root of
        # the extraction directory
        for file_path in file_path_map.values():
            file_path.rename(extraction_path / file_path.name)

        # remove any sub-directories within the extraction directory
        for path in extraction_path.iterdir():
            if path.is_dir():
                shutil.rmtree(path)
        extraction_completed = True
    finally:
        if not extraction_completed and remove_incomplete and extraction_path.is_dir():
            shutil.rmtree(extraction_path)
        if remove_zip and zip_path.exists():
            zip_path.unlink()


def upload_custom_embedder_model(files: Sequence[StrPath], name: str) -> None:
    """
    Upload a custom embedder model from either a zip file or a pair
    consisting of a pytorch_model.bin file and a config.json file.

    Parameters
    ----------
    files : Sequence[StrPath]
        Paths to the files to upload.
    name : str
        The name to give to the uploaded custom embedder model.

    Raises
    ------
    NotProvidedError
        If no name or file paths are provided.
    UploadTypeError
        If a single uploaded file is not a .zip file or two uploaded
        files are not named "pytorch_model.bin" and "config.json".
    UploadLimitError
        If more than two file paths are provided.

    """
    if not files:
        raise NotProvidedError(entity=Entity.FILES, ui_msg=UIMessage.NO_UPLOADED_FILES)
    model_dir_path = validate_model(
        name,
        Entity.CUSTOM_EMBEDDER_MODEL,
        mode="not_exists",
    )
    sorted_file_paths = sorted([Path(f) for f in files], key=lambda f: f.suffix)
    match sorted_file_paths:
        case [file_path]:
            if zipfile.is_zipfile(file_path):
                _extract_custom_embedder_model(file_path, model_dir_path)
            else:
                raise UploadTypeError(
                    entity=Entity.FILES,
                    valid_types=[".zip"],
                    type_class="formats",
                    multiple=False,
                )
        case [bin_path, json_path]:
            if bin_path.name == "pytorch_model.bin" and json_path.name == "config.json":
                copy_files_to_new_dir([bin_path, json_path], model_dir_path)
            else:
                raise UploadTypeError(
                    entity=Entity.FILES,
                    valid_types=["pytorch_model.bin", "config.json"],
                    type_class="names",
                    multiple=True,
                )
        case _:
            raise UploadLimitError(entity=Entity.FILES, limit="two")


def delete_models(
    directory: StrPath,
    names: Sequence[str],
    entity: ModelEntity = Entity.MODEL,
    ui_msg: UIMessage = UIMessage.NO_MODELS,
) -> None:
    """
    Delete the models with the provided names.

    Parameters
    ----------
    directory : StrPath
        The path to the directory containing the models to delete.
    names : Sequence[str]
        Names of the models to delete.
    entity : ModelEntity, optional
        The model entity being deleted.
    ui_msg : UIMessage, optional
        The message to display if no model names are provided.

    Raises
    ------
    NotProvidedError
        If no names of items are provided.
    ModelNotFoundError
        if an item with a provided name is not found.

    """
    if not names:
        raise NotProvidedError(entity=Entity.MODEL_NAMES, ui_msg=ui_msg)
    model_dir_paths: list[Path] = []
    for name in names:
        model_dir_path = Path(directory) / name
        if not model_dir_path.is_dir():
            raise ModelNotFoundError(entity=entity, name=name)
        model_dir_paths.append(model_dir_path)
    for model_dir_path in model_dir_paths:
        shutil.rmtree(model_dir_path)


def delete_voice_models(names: Sequence[str]) -> None:
    """
    Delete one or more voice models.

    Parameters
    ----------
    names : Sequence[str]
        Names of the voice models to delete.

    """
    delete_models(
        VOICE_MODELS_DIR,
        names,
        entity=Entity.VOICE_MODEL,
        ui_msg=UIMessage.NO_VOICE_MODELS,
    )


def delete_custom_embedder_models(names: Sequence[str]) -> None:
    """
    Delete one or more custom embedder models.

    Parameters
    ----------
    names : Sequence[str]
        Names of the custom embedder models to delete.

    """
    delete_models(
        CUSTOM_EMBEDDER_MODELS_DIR,
        names,
        entity=Entity.CUSTOM_EMBEDDER_MODEL,
        ui_msg=UIMessage.NO_CUSTOM_EMBEDDER_MODELS,
    )


def delete_custom_pretrained_models(names: Sequence[str]) -> None:
    """
    Delete one or more custom pretrained models.

    Parameters
    ----------
    names : Sequence[str]
        Names of the custom pretrained models to delete.

    """
    delete_models(
        CUSTOM_PRETRAINED_MODELS_DIR,
        names,
        entity=Entity.CUSTOM_PRETRAINED_MODEL,
        ui_msg=UIMessage.NO_CUSTOM_PRETRAINED_MODELS,
    )


def delete_training_models(names: Sequence[str]) -> None:
    """
    Delete one or more training models.

    Parameters
    ----------
    names : Sequence[str]
        Names of the training models to delete.

    """
    delete_models(
        TRAINING_MODELS_DIR,
        names,
        entity=Entity.TRAINING_MODEL,
        ui_msg=UIMessage.NO_TRAINING_MODELS,
    )


def delete_all_voice_models() -> None:
    """Delete all voice models."""
    delete_directory(VOICE_MODELS_DIR)


def delete_all_custom_embedder_models() -> None:
    """Delete all custom embedder models."""
    delete_directory(CUSTOM_EMBEDDER_MODELS_DIR)


def delete_all_custom_pretrained_models() -> None:
    """Delete all custom pretrained models."""
    delete_directory(CUSTOM_PRETRAINED_MODELS_DIR)


def delete_all_training_models() -> None:
    """Delete all training models."""
    delete_directory(TRAINING_MODELS_DIR)


def delete_all_models() -> None:
    """Delete all voice and training models."""
    display_progress("[~] Deleting all voice models ...", 0.0)
    delete_all_voice_models()
    display_progress("[~] Deleting all custom embedder models ...", 0.25)
    delete_all_custom_embedder_models()
    display_progress("[~] Deleting all custom pretrained models ...", 0.5)
    delete_all_custom_pretrained_models()
    display_progress("[~] Deleting all training models ...", 0.75)
    delete_all_training_models()



"""
Module which defines the command-line interface for generating
audio using RVC.
"""

from __future__ import annotations

from typing import Annotated

import time

# NOTE typer actually uses Path from pathlib at runtime
# even though it appears it is only a type annotation
from pathlib import Path  # noqa: TC003

import typer
from rich import print as rprint
from rich.panel import Panel

from ultimate_rvc.cli.common import (
    complete_audio_ext,
    complete_embedder_model,
    complete_f0_method,
    format_duration,
)
from ultimate_rvc.cli.generate.song_cover import app as song_cover_app
from ultimate_rvc.cli.generate.speech import app as speech_app
from ultimate_rvc.cli.typing_extra import PanelName
from ultimate_rvc.core.generate.common import convert as _convert
from ultimate_rvc.core.generate.common import wavify as _wavify
from ultimate_rvc.typing_extra import AudioExt, EmbedderModel, F0Method, RVCContentType

app = typer.Typer(
    name="generate",
    no_args_is_help=True,
    help="Generate audio using RVC",
    rich_markup_mode="markdown",
)


app.add_typer(song_cover_app)
app.add_typer(speech_app)


@app.command(no_args_is_help=True)
def wavify(
    audio_track: Annotated[
        Path,
        typer.Argument(
            help="The path to the audio track to convert.",
            exists=True,
            file_okay=True,
            dir_okay=False,
            resolve_path=True,
        ),
    ],
    directory: Annotated[
        Path,
        typer.Argument(
            help=(
                "The path to the directory where the converted audio track will be"
                " saved."
            ),
            exists=True,
            file_okay=False,
            dir_okay=True,
            resolve_path=True,
        ),
    ],
    prefix: Annotated[
        str,
        typer.Argument(
            help="The prefix to use for the name of the converted audio track.",
        ),
    ],
    accepted_format: Annotated[
        list[AudioExt] | None,
        typer.Option(
            case_sensitive=False,
            autocompletion=complete_audio_ext,
            help=(
                "An audio format to accept for conversion. This option can be provided"
                " multiple times to accept multiple formats. If not provided, the"
                " default accepted formats are mp3, ogg, flac, m4a and aac."
            ),
        ),
    ] = None,
) -> None:
    """
    Convert an audio track to wav format if its current format is an
    accepted format.
    """
    start_time = time.perf_counter()

    rprint()

    wav_path = _wavify(
        audio_track=audio_track,
        directory=directory,
        prefix=prefix,
        accepted_formats=set(accepted_format) if accepted_format else None,
    )
    if wav_path == audio_track:
        rprint(
            "[+] Audio track was not converted to WAV format. Presumably, "
            "its format is not in the given list of accepted formats.",
        )
    else:
        rprint("[+] Audio track succesfully converted to WAV format!")
        rprint()
        rprint("Elapsed time:", format_duration(time.perf_counter() - start_time))
        rprint(Panel(f"[green]{wav_path}", title="WAV Audio Track Path"))


@app.command(no_args_is_help=True)
def convert_voice(
    voice_track: Annotated[
        Path,
        typer.Argument(
            help="The path to the voice track to convert.",
            exists=True,
            file_okay=True,
            dir_okay=False,
            resolve_path=True,
        ),
    ],
    directory: Annotated[
        Path,
        typer.Argument(
            help=(
                "The path to the directory where the converted voice track will be"
                " will be saved."
            ),
            exists=True,
            file_okay=False,
            dir_okay=True,
            resolve_path=True,
        ),
    ],
    model_name: Annotated[
        str,
        typer.Argument(
            help="The name of the model to use for voice conversion.",
        ),
    ],
    n_octaves: Annotated[
        int,
        typer.Option(
            rich_help_panel=PanelName.MAIN_OPTIONS,
            help=(
                "The number of octaves to pitch-shift the converted voice by. Use"
                " 1 for male-to-female and -1 for vice-versa."
            ),
        ),
    ] = 0,
    n_semitones: Annotated[
        int,
        typer.Option(
            rich_help_panel=PanelName.MAIN_OPTIONS,
            help=(
                "The number of semi-tones to pitch-shift the converted"
                " voice by. Altering this slightly reduces sound quality."
            ),
        ),
    ] = 0,
    f0_method: Annotated[
        list[F0Method] | None,
        typer.Option(
            case_sensitive=False,
            autocompletion=complete_f0_method,
            rich_help_panel=PanelName.VOICE_SYNTHESIS_OPTIONS,
            help=(
                "The method to use for pitch extraction. This"
                " option can be provided multiple times to use multiple pitch"
                " extraction methods in combination. If not provided, will default to"
                " the rmvpe method, which is generally recommended."
            ),
        ),
    ] = None,
    index_rate: Annotated[
        float,
        typer.Option(
            min=0,
            max=1,
            rich_help_panel=PanelName.VOICE_SYNTHESIS_OPTIONS,
            help=(
                "The rate of influence of the index file. Increase to bias the"
                " conversion towards the accent of the voice model. Decrease to"
                " potentially reduce artifacts."
            ),
        ),
    ] = 0.3,
    rms_mix_rate: Annotated[
        float,
        typer.Option(
            min=0,
            max=1,
            rich_help_panel=PanelName.VOICE_SYNTHESIS_OPTIONS,
            help=(
                "Blending rate for the volume envelope of the converted voice. Controls"
                " how much to mimic the loudness of the input voice (0) or a fixed"
                " loudness (1)."
            ),
        ),
    ] = 1.0,
    protect_rate: Annotated[
        float,
        typer.Option(
            min=0,
            max=0.5,
            rich_help_panel=PanelName.VOICE_SYNTHESIS_OPTIONS,
            help=(
                "A coefficient which controls the extent to which consonants and"
                " breathing sounds are protected from artifacts. A higher value"
                " offers more protection but may worsen the indexing"
                " effect."
            ),
        ),
    ] = 0.33,
    hop_length: Annotated[
        int,
        typer.Option(
            min=1,
            max=512,
            rich_help_panel=PanelName.VOICE_SYNTHESIS_OPTIONS,
            help=(
                "Controls how often the CREPE-based pitch extraction method checks for"
                " pitch changes measured in milliseconds. Lower values lead to longer"
                " conversion times and a higher risk of voice cracks, but better pitch"
                " accuracy."
            ),
        ),
    ] = 128,
    split_voice: Annotated[
        bool,
        typer.Option(
            rich_help_panel=PanelName.VOICE_ENRICHMENT_OPTIONS,
            help=(
                "Whether to split the voice track into smaller segments"
                " before converting it. This can improve output quality for"
                " longer voice tracks."
            ),
        ),
    ] = False,
    autotune_voice: Annotated[
        bool,
        typer.Option(
            rich_help_panel=PanelName.VOICE_ENRICHMENT_OPTIONS,
            help="Whether to apply autotune to the converted voice.",
        ),
    ] = False,
    autotune_strength: Annotated[
        float,
        typer.Option(
            min=0,
            max=1,
            rich_help_panel=PanelName.VOICE_ENRICHMENT_OPTIONS,
            help=(
                "The intensity of the autotune effect to apply to the converted voice."
                " Higher values result in stronger snapping to the chromatic grid."
            ),
        ),
    ] = 1.0,
    clean_voice: Annotated[
        bool,
        typer.Option(
            rich_help_panel=PanelName.VOICE_ENRICHMENT_OPTIONS,
            help=(
                "Whether to clean the converted voice using noise reduction algorithms"
            ),
        ),
    ] = False,
    clean_strength: Annotated[
        float,
        typer.Option(
            min=0,
            max=1,
            rich_help_panel=PanelName.VOICE_ENRICHMENT_OPTIONS,
            help=(
                "The intensity of the cleaning to apply to the converted voice. Higher"
                " values result in stronger cleaning, but may lead to a more compressed"
                " sound."
            ),
        ),
    ] = 0.7,
    embedder_model: Annotated[
        EmbedderModel,
        typer.Option(
            case_sensitive=False,
            autocompletion=complete_embedder_model,
            rich_help_panel=PanelName.SPEAKER_EMBEDDINGS_OPTIONS,
            help="The model to use for generating speaker embeddings.",
        ),
    ] = EmbedderModel.CONTENTVEC,
    custom_embedder_model: Annotated[
        str | None,
        typer.Option(
            rich_help_panel=PanelName.SPEAKER_EMBEDDINGS_OPTIONS,
            help=(
                "The name of a custom embedder model to use for generating speaker"
                " embeddings. Only applicable if `embedder-model` is set to `custom`."
            ),
        ),
    ] = None,
    sid: Annotated[
        int,
        typer.Option(
            rich_help_panel=PanelName.SPEAKER_EMBEDDINGS_OPTIONS,
            help="The id of the speaker to use for multi-speaker RVC models.",
        ),
    ] = 0,
) -> None:
    """Convert a voice track using RVC."""
    start_time = time.perf_counter()

    rprint()

    converted_voice_path = _convert(
        audio_track=voice_track,
        directory=directory,
        model_name=model_name,
        n_octaves=n_octaves,
        n_semitones=n_semitones,
        f0_methods=f0_method,
        index_rate=index_rate,
        rms_mix_rate=rms_mix_rate,
        protect_rate=protect_rate,
        hop_length=hop_length,
        split_audio=split_voice,
        autotune_audio=autotune_voice,
        autotune_strength=autotune_strength,
        clean_audio=clean_voice,
        clean_strength=clean_strength,
        embedder_model=embedder_model,
        custom_embedder_model=custom_embedder_model,
        sid=sid,
        content_type=RVCContentType.VOICE,
    )
    rprint("[+] Voice track succesfully converted!")
    rprint()
    rprint("Elapsed time:", format_duration(time.perf_counter() - start_time))
    rprint(Panel(f"[green]{converted_voice_path}", title="Converted Voice Path"))



import os
import re
import shutil
import sys
import zipfile
from urllib.parse import unquote

import six
from bs4 import BeautifulSoup

import requests

from tqdm import tqdm

now_dir = os.getcwd()
sys.path.append(now_dir)

from ultimate_rvc.rvc.lib.tools import gdown
from ultimate_rvc.rvc.lib.utils import format_title

file_path = os.path.join(now_dir, "logs")
zips_path = os.path.join(file_path, "zips")
os.makedirs(zips_path, exist_ok=True)


def search_pth_index(folder):
    pth_paths = [
        os.path.join(folder, file)
        for file in os.listdir(folder)
        if os.path.isfile(os.path.join(folder, file)) and file.endswith(".pth")
    ]
    index_paths = [
        os.path.join(folder, file)
        for file in os.listdir(folder)
        if os.path.isfile(os.path.join(folder, file)) and file.endswith(".index")
    ]
    return pth_paths, index_paths


def download_from_url(url):
    os.chdir(zips_path)

    try:
        if "drive.google.com" in url:
            file_id = extract_google_drive_id(url)
            if file_id:
                gdown.download(
                    url=f"https://drive.google.com/uc?id={file_id}",
                    quiet=False,
                    fuzzy=True,
                )
        elif "/blob/" in url or "/resolve/" in url:
            download_blob_or_resolve(url)
        elif "/tree/main" in url:
            download_from_huggingface(url)
        else:
            download_file(url)

        rename_downloaded_files()
        return "downloaded"
    except Exception as error:
        print(f"An error occurred downloading the file: {error}")
        return None
    finally:
        os.chdir(now_dir)


def extract_google_drive_id(url):
    if "file/d/" in url:
        return url.split("file/d/")[1].split("/")[0]
    if "id=" in url:
        return url.split("id=")[1].split("&")[0]
    return None


def download_blob_or_resolve(url):
    if "/blob/" in url:
        url = url.replace("/blob/", "/resolve/")
    response = requests.get(url, stream=True)
    if response.status_code == 200:
        save_response_content(response)
    else:
        raise ValueError(
            "Download failed with status code: " + str(response.status_code),
        )


def save_response_content(response):
    content_disposition = unquote(response.headers.get("Content-Disposition", ""))
    file_name = (
        re.search(r'filename="([^"]+)"', content_disposition)
        .groups()[0]
        .replace(os.path.sep, "_")
        if content_disposition
        else "downloaded_file"
    )

    total_size = int(response.headers.get("Content-Length", 0))
    chunk_size = 1024

    with (
        open(os.path.join(zips_path, file_name), "wb") as file,
        tqdm(
            total=total_size,
            unit="B",
            unit_scale=True,
            desc=file_name,
        ) as progress_bar,
    ):
        for data in response.iter_content(chunk_size):
            file.write(data)
            progress_bar.update(len(data))


def download_from_huggingface(url):
    response = requests.get(url)
    soup = BeautifulSoup(response.content, "html.parser")
    temp_url = next(
        (
            link["href"]
            for link in soup.find_all("a", href=True)
            if link["href"].endswith(".zip")
        ),
        None,
    )
    if temp_url:
        url = temp_url.replace("blob", "resolve")
        if "huggingface.co" not in url:
            url = "https://huggingface.co" + url
        download_file(url)
    else:
        raise ValueError("No zip file found in Huggingface URL")


def download_file(url):
    response = requests.get(url, stream=True)
    if response.status_code == 200:
        save_response_content(response)
    else:
        raise ValueError(
            "Download failed with status code: " + str(response.status_code),
        )


def rename_downloaded_files():
    for currentPath, _, zipFiles in os.walk(zips_path):
        for file in zipFiles:
            file_name, extension = os.path.splitext(file)
            real_path = os.path.join(currentPath, file)
            os.rename(real_path, file_name.replace(os.path.sep, "_") + extension)


def extract(zipfile_path, unzips_path):
    try:
        with zipfile.ZipFile(zipfile_path, "r") as zip_ref:
            zip_ref.extractall(unzips_path)
        os.remove(zipfile_path)
        return True
    except Exception as error:
        print(f"An error occurred extracting the zip file: {error}")
        return False


def unzip_file(zip_path, zip_file_name):
    zip_file_path = os.path.join(zip_path, zip_file_name + ".zip")
    extract_path = os.path.join(file_path, zip_file_name)
    with zipfile.ZipFile(zip_file_path, "r") as zip_ref:
        zip_ref.extractall(extract_path)
    os.remove(zip_file_path)


def model_download_pipeline(url: str):
    try:
        result = download_from_url(url)
        if result == "downloaded":
            return handle_extraction_process()
        return "Error"
    except Exception as error:
        print(f"An unexpected error occurred: {error}")
        return "Error"


def handle_extraction_process():
    extract_folder_path = ""
    for filename in os.listdir(zips_path):
        if filename.endswith(".zip"):
            zipfile_path = os.path.join(zips_path, filename)
            model_name = format_title(os.path.basename(zipfile_path).split(".zip")[0])
            extract_folder_path = os.path.join("logs", os.path.normpath(model_name))
            success = extract(zipfile_path, extract_folder_path)
            clean_extracted_files(extract_folder_path, model_name)

            if success:
                print(f"Model {model_name} downloaded!")
            else:
                print(f"Error downloading {model_name}")
                return "Error"
    if not extract_folder_path:
        print("Zip file was not found.")
        return "Error"
    return search_pth_index(extract_folder_path)


def clean_extracted_files(extract_folder_path, model_name):
    macosx_path = os.path.join(extract_folder_path, "__MACOSX")
    if os.path.exists(macosx_path):
        shutil.rmtree(macosx_path)

    subfolders = [
        f
        for f in os.listdir(extract_folder_path)
        if os.path.isdir(os.path.join(extract_folder_path, f))
    ]
    if len(subfolders) == 1:
        subfolder_path = os.path.join(extract_folder_path, subfolders[0])
        for item in os.listdir(subfolder_path):
            shutil.move(
                os.path.join(subfolder_path, item),
                os.path.join(extract_folder_path, item),
            )
        os.rmdir(subfolder_path)

    for item in os.listdir(extract_folder_path):
        source_path = os.path.join(extract_folder_path, item)
        if ".pth" in item:
            new_file_name = model_name + ".pth"
        elif ".index" in item:
            new_file_name = model_name + ".index"
        else:
            continue

        destination_path = os.path.join(extract_folder_path, new_file_name)
        if not os.path.exists(destination_path):
            os.rename(source_path, destination_path)



"""
Module defining common utility functions and classes for the
web application of the Ultimate RVC project.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Concatenate, ParamSpec, TypeVar

from functools import partial

import gradio as gr

from ultimate_rvc.core.exceptions import NotProvidedError
from ultimate_rvc.core.manage.config import load_config, save_config
from ultimate_rvc.web.config.main import TotalConfig
from ultimate_rvc.web.typing_extra import SongTransferOption, SpeechTransferOption

if TYPE_CHECKING:
    from collections.abc import Callable, Sequence

    from gradio.components import Component
    from gradio.events import Dependency

    from ultimate_rvc.web.config.component import AudioConfig
    from ultimate_rvc.web.typing_extra import (
        BaseDropdownChoices,
        BaseDropdownValue,
        ComponentVisibilityKwArgs,
        DropdownChoices,
        DropdownValue,
        TextBoxKwArgs,
        UpdateAudioKwArgs,
        UpdateDropdownKwArgs,
    )

PROGRESS_BAR = gr.Progress()

T = TypeVar("T")
P = ParamSpec("P")
U = TypeVar("U", bound=SongTransferOption | SpeechTransferOption)


def exception_harness(  # noqa: UP047
    fn: Callable[P, T],
    info_msg: str | None = None,
) -> Callable[P, T]:
    """
    Wrap a function in a harness that catches exceptions and re-raises
    them as instances of `gradio.Error`.

    Parameters
    ----------
    fn : Callable[P, T]
        The function to wrap.

    info_msg : str, optional
        Message to display in an info-box pop-up after the function
        executes successfully.

    Returns
    -------
    Callable[P, T]
        The wrapped function.

    """

    def _wrapped_fn(*args: P.args, **kwargs: P.kwargs) -> T:
        try:
            res = fn(*args, **kwargs)
        except gr.Error:
            raise
        except NotProvidedError as e:
            msg = e.ui_msg or e
            raise gr.Error(str(msg)) from None
        except Exception as e:
            raise gr.Error(str(e)) from e
        else:
            if info_msg:
                gr.Success(info_msg, duration=1)
            return res

    return _wrapped_fn


def confirmation_harness(  # noqa: UP047
    fn: Callable[P, T],
) -> Callable[Concatenate[bool, P], T]:
    """
    Wrap a function in a harness that requires a confirmation before
    executing and catches exceptions, re-raising them as instances of
    `gradio.Error`.

    Parameters
    ----------
    fn : Callable[P, T]
        The function to wrap.

    Returns
    -------
    Callable[Concatenate[bool, P], T]
        The wrapped function.

    """

    def _wrapped_fn(confirm: bool, *args: P.args, **kwargs: P.kwargs) -> T:
        if confirm:
            return exception_harness(fn)(*args, **kwargs)
        err_msg = "Confirmation missing!"
        raise gr.Error(err_msg)

    return _wrapped_fn


def render_msg(
    template: str,
    *args: str,
    display_info: bool = False,
    **kwargs: str,
) -> str:
    """
    Render a message template with the provided arguments.

    Parameters
    ----------
    template : str
        Message template to render.
    args : str
        Positional arguments to pass to the template.
    display_info : bool, default=False
        Whether to display the rendered message as an info message
        in addition to returning it.
    kwargs : str
        Keyword arguments to pass to the template.

    Returns
    -------
    str
        Rendered message.

    """
    msg = template.format(*args, **kwargs)
    if display_info:
        gr.Info(msg)
    return msg


def confirm_box_js(msg: str) -> str:
    """
    Generate a JavaScript code snippet which:
      * defines an anonymous function that takes one named parameter and
      zero or more unnamed parameters
      * renders a confirmation box
      * returns the choice selected by the user in that confirmation
      box in addition to any unnamed parameters passed to the function.

    Parameters
    ----------
    msg : str
        Message to display in the confirmation box rendered by the
        JavaScript code snippet.

    Returns
    -------
    str
        The JavaScript code snippet.

    """
    return f"(x, ...args) => [confirm('{msg}'), ...args]"


def update_value(x: str | None) -> dict[str, Any]:
    """
    Update the value of a component.

    Parameters
    ----------
    x : str | None
        New value for the component.

    Returns
    -------
    dict[str, Any]
        Dictionary which updates the value of the component.

    """
    return gr.update(value=x)


def update_values(*xs: str) -> tuple[dict[str, Any], ...]:
    """
    Update the values of multiple components.

    Parameters
    ----------
    xs : str
        New values for the components.

    Returns
    -------
    tuple[dict[str, Any], ...]
        Tuple of dictionaries which update the values of the components.

    """
    return tuple(gr.update(value=x) for x in xs)


def toggle_visibility(
    value: T,
    targets: set[T],
    default: str | float | None = None,
    update_default: bool = False,
) -> dict[str, Any]:
    """
    Toggle the visibility of a component based on equality of
    a value and one of a set of targets.

    Parameters
    ----------
    value : T
        The value to compare against the target.
    targets : set[T]
        The set of targets to compare the value against.
    default : str | float | None, optional
        Default value for the component.
    update_default : bool, default=False
        Whether to update the default value of the component.

    Returns
    -------
    dict[str, Any]
        Dictionary which updates the visibility of the component.

    """
    update_args: ComponentVisibilityKwArgs = {"visible": value in targets}
    if update_default:
        update_args["value"] = default
    return gr.update(**update_args)


def toggle_visibilities(
    num_components: int,
    value: T,
    targets: set[T],
    defaults: Sequence[str | float | None] = [],
    update_default: bool = False,
) -> list[dict[str, Any]]:
    """
    Toggle the visibility of multiple components based on equality of
    a value and one of a set of targets.

    Parameters
    ----------
    num_components : int
        Number of components to set visibility for.
    value : T
        The value to compare against the target.
    targets : set[T]
        The set of targets to compare the value against.
    defaults : Sequence[str | float | None], optional
        Default values for the components.
    update_default : bool, default=False
        whether to update the default values of the components.

    Returns
    -------
    list[dict[str, Any]]
        List of dictionaries which update the visibility of the
        components.

    Raises
    ------
    ValueError
        If the number of default values does not match the number of
        components.

    """
    if update_default and len(defaults) != num_components:
        err_msg = "Number of default values must be equal to the number of components."
        raise ValueError(err_msg)
    update_args_list: list[ComponentVisibilityKwArgs] = []
    for index in range(num_components):
        update_args: ComponentVisibilityKwArgs = {"visible": value in targets}
        if update_default:
            update_args["value"] = defaults[index]
        update_args_list.append(update_args)
    return [gr.update(**update_args) for update_args in update_args_list]


def toggle_visible_component(
    num_components: int,
    visible_index: int,
    reset_values: bool = True,
) -> dict[str, Any] | tuple[dict[str, Any], ...]:
    """
    Reveal a single component from a set of components. All other
    components are hidden.

    Parameters
    ----------
    num_components : int
        Number of components to set visibility for.
    visible_index : int
        Index of the component to reveal.
    reset_values : bool, default=True
        Whether to reset the values of the components.

    Returns
    -------
    dict[str, Any] | tuple[dict[str, Any], ...]
        A single dictionary or a tuple of dictionaries that update the
        visibility of the components.

    Raises
    ------
    ValueError
        If the visible index exceeds or is equal to the number of
        components to set visibility for.

    """
    if visible_index >= num_components:
        err_msg = (
            "Visible index must be less than the number of components to set visibility"
            " for."
        )
        raise ValueError(err_msg)
    update_args_list: list[ComponentVisibilityKwArgs] = []
    for _ in range(num_components):
        update_args: ComponentVisibilityKwArgs = {"visible": False}
        if reset_values:
            update_args["value"] = None
        update_args_list.append(update_args)
    update_args_list[visible_index]["visible"] = True
    match update_args_list:
        case [update_args]:
            return gr.update(**update_args)
        case _:
            return tuple(gr.update(**update_args) for update_args in update_args_list)


def initialize_dropdowns(
    fn: Callable[P, BaseDropdownChoices],
    num_components: int,
    value: BaseDropdownValue = None,
    value_indices: Sequence[int] = [],
    *args: P.args,
    **kwargs: P.kwargs,
) -> list[gr.Dropdown]:
    """
    Initialize the choices and optionally the value of one or more
    dropdown components.


    Parameters
    ----------
    fn : Callable[P, BaseDropdownChoices]
        Function to get initial choices for the dropdown components.
    num_components : int
        Number of dropdown components to initialize.
    value : BaseDropdownValue, optional
        New value for dropdown components.
    value_indices : Sequence[int], default=[]
        Indices of dropdown components to initialize the value for.
    args : P.args
        Positional arguments to pass to the function used to initialize
        dropdown choices.
    kwargs : P.kwargs
        Keyword arguments to pass to the function used to initialize
        dropdown choices.

    Returns
    -------
    list[gr.Dropdown]
        List of initialized dropdown components.

    Raises
    ------
    ValueError
        If not all provided indices are unique or if an index exceeds
        or is equal to the number of dropdown components.

    """
    if len(value_indices) != len(set(value_indices)):
        err_msg = "Value indices must be unique."
        raise ValueError(err_msg)
    if value_indices and max(value_indices) >= num_components:
        err_msg = (
            "Index of a dropdown component to update the value for exceeds the number"
            " of dropdown components to update."
        )
        raise ValueError(err_msg)
    updated_choices = fn(*args, **kwargs)
    if value is None or value not in updated_choices:
        value = next(iter(updated_choices), None)
        if isinstance(value, tuple):
            value = value[1]
    update_args_list: list[UpdateDropdownKwArgs] = [
        {"choices": updated_choices} for _ in range(num_components)
    ]
    for index in value_indices:
        update_args_list[index]["value"] = value
    return [gr.Dropdown(**update_args) for update_args in update_args_list]


def update_dropdowns(
    fn: Callable[P, DropdownChoices],
    num_components: int,
    value: DropdownValue = None,
    value_indices: Sequence[int] = [],
    *args: P.args,
    **kwargs: P.kwargs,
) -> gr.Dropdown | tuple[gr.Dropdown, ...]:
    """
    Update the choices and optionally the value of one or more dropdown
    components.

    Parameters
    ----------
    fn : Callable[P, DropdownChoices]
        Function to get updated choices for the dropdown components.
    num_components : int
        Number of dropdown components to update.
    value : DropdownValue, optional
        New value for dropdown components.
    value_indices : Sequence[int], default=[]
        Indices of dropdown components to update the value for.
    args : P.args
        Positional arguments to pass to the function used to update
        dropdown choices.
    kwargs : P.kwargs
        Keyword arguments to pass to the function used to update
        dropdown choices.

    Returns
    -------
    gr.Dropdown | tuple[gr.Dropdown,...]
        Updated dropdown component or components.

    Raises
    ------
    ValueError
        If not all provided indices are unique or if an index exceeds
        or is equal to the number of dropdown components.

    """
    if len(value_indices) != len(set(value_indices)):
        err_msg = "Value indices must be unique."
        raise ValueError(err_msg)
    if value_indices and max(value_indices) >= num_components:
        err_msg = (
            "Index of a dropdown component to update the value for exceeds the number"
            " of dropdown components to update."
        )
        raise ValueError(err_msg)
    updated_choices = fn(*args, **kwargs)
    update_args_list: list[UpdateDropdownKwArgs] = [
        {"choices": updated_choices} for _ in range(num_components)
    ]
    for index in value_indices:
        update_args_list[index]["value"] = value

    match update_args_list:
        case [update_args]:
            # NOTE This is a workaround as gradio does not support
            # singleton tuples for components.
            return gr.Dropdown(**update_args)
        case _:
            return tuple(gr.Dropdown(**update_args) for update_args in update_args_list)


def toggle_intermediate_audio(visible: bool, num_components: int) -> list[gr.Accordion]:
    """
    Toggle the visibility of intermediate audio accordions.

    Parameters
    ----------
    visible : bool
        Visibility status of the intermediate audio accordions.

    num_components : int
        Number of intermediate audio accordions to toggle visibility
        for.

    Returns
    -------
    list[gr.Accordion]
        The intermediate audio accordions.

    """
    accordions = [gr.Accordion(open=False) for _ in range(num_components)]
    return [gr.Accordion(visible=visible, open=False), *accordions]


def update_output_name(
    fn: Callable[P, str],
    update_placeholder: bool = False,
    *args: P.args,
    **kwargs: P.kwargs,
) -> gr.Textbox:
    """
    Update a textbox component so that it displays a suitable name for
    an output audio file.

    Parameters
    ----------
    fn : Callable[..., str]
        Function to generate the output name.
    update_placeholder : bool, default=False
        Whether to update the placeholder text instead of the value of
        the textbox component.
    args : P.args
        Positional arguments to pass to the function used to generate
        the output name.
    kwargs : P.kwargs
        Keyword arguments to pass to the function used to generate the
        output name.

    Returns
    -------
    gr.Textbox
        Textbox component with updated value or placeholder text.

    """
    update_args: TextBoxKwArgs = {}
    update_key = "placeholder" if update_placeholder else "value"
    if (args and any(args)) or (kwargs and any(kwargs)):
        output_name = fn(*args, **kwargs)
        update_args[update_key] = output_name
    else:
        update_args[update_key] = None
    return gr.Textbox(**update_args)


def update_audio(
    num_components: int,
    output_indices: Sequence[int],
    track: str | None,
    disallow_none: bool = True,
) -> gr.Audio | tuple[gr.Audio, ...]:
    """
    Update the value of a subset of `Audio` components to the given
    audio track.

    Parameters
    ----------
    num_components : int
        The total number of `Audio` components under consideration.
    output_indices : Sequence[int]
        Indices of `Audio` components to update the value for.
    track : str
        Path pointing to an audio track to update the value of the
        indexed `Audio` components with.
    disallow_none : bool, default=True
        Whether to disallow the value of the indexed components to be
        `None`.

    Returns
    -------
    gr.Audio | tuple[gr.Audio, ...]
        Each `Audio` component under consideration with the value of the
        indexed components updated to the given audio track.

    """
    update_args_list: list[UpdateAudioKwArgs] = [{} for _ in range(num_components)]
    for index in output_indices:
        if track or not disallow_none:
            update_args_list[index]["value"] = track
    match update_args_list:
        case [update_args]:
            return gr.Audio(**update_args)
        case _:
            return tuple(gr.Audio(**update_args) for update_args in update_args_list)


def render_transfer_component(
    value: list[U],
    label_prefix: str,
    option_type: type[U],
) -> gr.Dropdown:
    """
    Render a dropdown for transferring tracks.

    Parameters
    ----------
    value : list[U]
        The default selected values for the dropdown.
    label_prefix : str
        The prefix for the dropdown label.
    option_type : type[U]
        The type of the transfer options.

    Returns
    -------
    gr.Dropdown
        A dropdown component for transferring tracks.

    """
    return gr.Dropdown(
        choices=list(option_type),
        label=f"{label_prefix} destination",
        info=(
            "Select the input track(s) to transfer the"
            f" {label_prefix.lower()} to when the 'Transfer"
            f" {label_prefix.lower()}' button is clicked."
        ),
        type="index",
        multiselect=True,
        value=value,
    )


def setup_transfer_event(
    btn: gr.Button,
    transfer: gr.Dropdown,
    output: gr.Audio,
    input_configs: Sequence[AudioConfig],
) -> Dependency:
    """
    Set up a transfer track event for a button component.


    Parameters
    ----------
    btn : gr.Button
        The button to set up the event for.
    transfer : gr.Dropdown
        A dropdown component containing transfer options.
    output : gr.Audio
        An audio component containing the track to transfer.
    input_configs : Sequence[AudioConfig]
        A sequence of configurations for all input audio components
        that can be transferred to.

    Returns
    -------
    Dependency
        The event listener for the button click.


    """
    return btn.click(
        partial(update_audio, len(input_configs)),
        inputs=[transfer, output],
        outputs=[c.instance for c in input_configs],
        show_progress="hidden",
    )


def setup_delete_event(
    button: gr.Button,
    fn: Callable[..., None],
    inputs: list[Component],
    outputs: gr.Textbox,
    confirm_msg: str,
    success_msg: str,
) -> Dependency:
    """
    Set up a delete event for a button component.

    Parameters
    ----------
    button : gr.Button
        Button component to set up the delete event for.
    fn : Callable[..., None]
        Function to call when the button is clicked.
    inputs : list[Component]
        List of input components to pass to the function that is called
        when the button is clicked.
    outputs : gr.Textbox
        Textbox component to update with a success message upon
        successful deletion.
    confirm_msg : str
        Message to display in a confirmation box before executing the
        delete event.
    success_msg : str
        Message to display in the provided textbox component upon
        successful deletion.


    Returns
    -------
    Dependency
        A dependency that can be used for sequencing the delete
        event with other events.

    """
    return button.click(
        confirmation_harness(fn),
        inputs=inputs,
        outputs=outputs,
        js=confirm_box_js(confirm_msg),
    ).success(partial(render_msg, success_msg), outputs=outputs, show_progress="hidden")


def save_total_config_values(name: str, *values: *tuple[Any, ...]) -> None:
    """
    Save the provided component values to a total configuration model
    and write it to a JSON file with the provided name.

    Parameters
    ----------
    name : str
        The name of the JSON file to write the total configuration model
        to.
    *values : *tuple[Any, ...]
        The component values to save to the total configuration model.

    """
    new_config = TotalConfig()
    for value, component_config in zip(values, new_config.all, strict=True):
        component_config.value = value

    save_config(name, new_config)


def load_total_config_values(name: str) -> tuple[Any, ...]:
    """
    Load a total configuration model from a JSON file with the provided
    name and return the non-excluded values of its nested component
    configurations.


    Parameters
    ----------
    name : str
        The name of the JSON file to load the total configuration model
        from.

    Returns
    -------
    tuple[Any, ...]
        The non-excluded values of the nested component configurations
        in the loaded total configuration model.

    """
    new_config = load_config(name, TotalConfig)
    return tuple(c.value for c in new_config.all)
